# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.

import gc
import multiprocessing
import os
import time
from typing import Union

import numpy as np
from numpy.lib import recfunctions as rfn

from remotior_sensus.core import configurations as cfg, messages, processor
from remotior_sensus.util import files_directories, raster_vector

try:
    multiprocessing.set_start_method('spawn')
except Exception as error:
    str(error)


# class for multiprocess functions
class Multiprocess(object):

    def __init__(self, n_processes: int = None):
        self.pool = multiprocessing.Pool(processes=n_processes)
        self.manager = multiprocessing.Manager()
        self.n_processes = n_processes
        self.output = False

    # start multiprocess
    def start(self, n_processes):
        self.stop()
        self.pool = multiprocessing.Pool(processes=n_processes)
        cfg.multiprocess = self
        self.manager = multiprocessing.Manager()
        self.n_processes = n_processes
        self.output = False

    # stop multiprocess
    def stop(self):
        try:
            self.pool.close()
            self.pool.terminate()
        except Exception as err:
            str(err)

    # run multiprocess
    def run(
            self, raster_path, function=None, function_argument=None,
            function_variable=None, calculation_datatype=None,
            use_value_as_nodata=None, any_nodata_mask=True,
            output_raster_path=None, output_data_type=None,
            output_nodata_value=None, compress=False, compress_format='LZW',
            n_processes: int = None, available_ram: int = None,
            dummy_bands=1, output_band_number=1, boundary_size=None,
            unique_section=False, keep_output_array=False,
            keep_output_argument=False, delete_array=True, scale=None,
            offset=None, input_nodata_as_value=None,
            classification=False, classification_confidence=False,
            signature_raster=False, virtual_raster=False,
            multi_add_factors=None, separate_bands=False,
            progress_message=None, device=None,
            min_progress=None, max_progress=None
    ):
        """
        :param device: processing device 'cpu' or 'cuda' if available.
        :param classification_confidence: if True, write also additional
            classification confidence rasters as output.
        :param signature_raster: if True, write additional rasters for each
            spectral signature as output.
        :param raster_path: input path.
        :param multi_add_factors: list of multiplicative and additive factors.
        :param virtual_raster: if True, create virtual raster output.
        :param offset: integer number of output offset.
        :param scale: integer number of output scale.
        :param delete_array: if True delete output array.
        :param keep_output_argument: if True keep output argument for post processing.
        :param keep_output_array: if True keep output array for post processing.
        :param unique_section: if True consider the whole raster as unique section.
        :param dummy_bands: integer number of dummy bands to be counted for
            calculating block size
        :param available_ram: integer value of RAM in MB.
        :param any_nodata_mask:  True to apply the nodata where any input is
            nodata, False to apply nodata where all inputs are nodata, None not apply nodata
        :param input_nodata_as_value: consider nodata as value in calculation
        :param output_data_type: string of data type for output raster such as Float32 or Int16
        :param output_band_number: number of bands of the output
        :param output_nodata_value: output nodata value
        :param compress_format: string of format of raster compression
        :param compress: True to compress the output raster or False not to compress
        :param boundary_size: integer number of pixels used to extend the
            boundary of calculations
        :param output_raster_path: list of output path strings
        :param calculation_datatype: datatype use during calculation
        :param function_variable: list of variables for function
        :param function_argument: arguments of function
        :param n_processes: number of parallel processes.
        :param function: function name
        :param classification: if True, settings are defined for a classification output
        :param use_value_as_nodata: integer value as nodata in calculation
        :param separate_bands: if True, calculate a section for each raster range
        :param progress_message: progress message
        :param min_progress: minimum progress value
        :param max_progress: maximum progress value
        """  # noqa: E501
        cfg.logger.log.debug('multiprocess: %s' % function)
        self.output = False
        process_result = {}
        process_output_files = {}
        # unit of memory
        memory_unit = cfg.memory_unit_array_12
        # calculation data type
        if calculation_datatype is None:
            calc_datatype = [np.float32]
        else:
            calc_datatype = [calculation_datatype]
        if (calculation_datatype == np.uint32
                or calculation_datatype == np.int32):
            memory_unit = cfg.memory_unit_array_8
        elif (calculation_datatype == np.uint16
              or calculation_datatype == np.int16):
            memory_unit = cfg.memory_unit_array_4
        if output_nodata_value is None:
            output_nodata_value = cfg.nodata_val
        if output_data_type is None:
            output_data_type = cfg.raster_data_type
        if n_processes is None:
            n_processes = self.n_processes
        elif n_processes > self.n_processes:
            self.start(n_processes)
        if device == 'cuda':
            n_processes = 1
        if min_progress is None:
            min_progress = 0
        if max_progress is None:
            max_progress = 100
        # reserve progress for writing output
        max_progress = round(max_progress * 0.9)
        if max_progress < min_progress:
            max_progress = min_progress
        max_progress_write = round(max_progress * 1.1)
        if max_progress_write > 100:
            max_progress_write = 100
        if progress_message is None:
            progress_message = 'processing'
        if available_ram is None:
            available_ram = cfg.available_ram
        # calculate block size
        block_size = _calculate_block_size(
            raster_path, n_processes, memory_unit, dummy_bands,
            available_ram=available_ram
        )
        if block_size is not False:
            (raster_x_size, raster_y_size, block_size_x, block_size_y,
             list_range_x, list_range_y, tot_blocks,
             number_of_bands) = block_size
        else:
            cfg.logger.log.error('unable to get raster: %s' % raster_path)
            messages.error('unable to get raster: %s' % raster_path)
            return
        # compute raster pieces
        pieces = _compute_raster_pieces(
            raster_x_size, raster_y_size, block_size_x, block_size_y,
            list_range_y, n_processes, separate_bands=separate_bands,
            boundary_size=boundary_size, unique_section=unique_section
        )
        # progress queue
        p_mq = self.manager.Queue()
        results = []
        for p in range(len(pieces)):
            process_parameters = [p, cfg.temp, available_ram, cfg.gdal_path,
                                  p_mq, cfg.refresh_time, memory_unit]
            input_parameters = [[raster_path], calc_datatype, boundary_size,
                                pieces[p], [scale], [offset],
                                [use_value_as_nodata], None,
                                input_nodata_as_value, multi_add_factors,
                                dummy_bands]
            output_parameters = [[output_raster_path], [output_data_type],
                                 compress, compress_format, any_nodata_mask,
                                 [output_nodata_value], [output_band_number],
                                 keep_output_array, keep_output_argument]
            c = self.pool.apply_async(
                processor.function_initiator,
                args=(process_parameters, input_parameters, output_parameters,
                      function, [function_argument], [function_variable],
                      False, classification, classification_confidence,
                      signature_raster)
            )
            results.append([c, p])
        while True:
            # update progress
            p_r = []
            for r in results:
                p_r.append(r[0].ready())
            if all(p_r):
                break
            time.sleep(cfg.refresh_time)
            # progress message
            try:
                p_m_qp = p_mq.get(False)
                count_progress = int(p_m_qp[0])
                length = int(p_m_qp[1])
                progress = int(100 * count_progress / length)
                cfg.progress.update(
                    message=progress_message, step=count_progress,
                    steps=length, minimum=min_progress,
                    maximum=max_progress, percentage=progress
                )
            except Exception as err:
                str(err)
        for r in results:
            res = r[0].get()
            if classification:
                process_result[r[1]] = res[0]
                process_output_files[r[1]] = res[1]
            else:
                process_result[r[1]] = res[0]
                process_output_files[r[1]] = res[1]
            if cfg.logger.level < 20:
                cfg.logger.log.debug(res[3])
            # error
            if res[2] is not False:
                cfg.logger.log.error('multiprocess: %s' % res[2])
                messages.error('multiprocess: %s' % res[2])
                gc.collect()
                return
        gc.collect()
        if classification:
            multiprocess_result = self._collect_classification_results(
                process_result, output_raster_path,
                function_argument=function_argument, scale=scale,
                offset=offset, virtual_raster=virtual_raster,
                output_nodata_value=output_nodata_value,
                output_data_type=output_data_type, compress=compress,
                compress_format=compress_format,
                delete_array=delete_array, n_processes=n_processes,
                available_ram=available_ram, min_progress=max_progress,
                max_progress=max_progress_write
            )
        else:
            multiprocess_result = self._collect_results(
                process_result, process_output_files, output_raster_path,
                scale=scale, offset=offset, virtual_raster=virtual_raster,
                output_nodata_value=output_nodata_value,
                output_data_type=output_data_type, compress=compress,
                compress_format=compress_format, delete_array=delete_array,
                n_processes=n_processes, available_ram=available_ram,
                min_progress=max_progress, max_progress=max_progress_write
            )
        cfg.progress.update(percentage=False)
        cfg.logger.log.debug('end')
        self.output = multiprocess_result

    # separated process for each input
    def run_separated(
            self, raster_path_list, function=None, function_argument=None,
            function_variable=None, calculation_datatype=None,
            use_value_as_nodata=None, any_nodata_mask=True,
            output_raster_list=None, output_data_type=None,
            output_nodata_value=None, compress=False, compress_format='LZW',
            n_processes: int = None, available_ram: int = None,
            output_band_number_list=None, boundary_size=None,
            dummy_bands=0,
            keep_output_array=False, keep_output_argument=False,
            scale=None, offset=None, input_nodata_as_value=None,
            multi_add_factors=None, progress_message=None, min_progress=None,
            max_progress=None
    ):
        """
        :param multi_add_factors: list of multiplicative and additive factors
        :param offset: list integer number of output offset
        :param scale: list of integer number of output scale
        :param keep_output_argument: if True keep output argument for
            post-processing
        :param keep_output_array: if True keep output array for post-processing
        :param available_ram: integer value of RAM in MB
        :param any_nodata_mask:  True to apply the nodata where any input is
            nodata, False to apply nodata where all
            inputs are nodata, None not apply nodata
        :param input_nodata_as_value: consider nodata as value in calculation
        :param output_data_type: list of data type string for output raster
            such as Float32 or Int16
        :param output_band_number_list: list of number of bands of the output
        :param output_nodata_value: nodata value of the output
        :param compress_format: string of format of raster compression
        :param compress: True to compress the output raster or False not to
            compress
        :param boundary_size: integer number of pixels used to extend the
            boundary of calculations
        :param output_raster_list: list of output path strings
        :param calculation_datatype: datatype use during calculation
        :param function_variable: list of variables for function
        :param function_argument: arguments of function
        :param n_processes: number of parallel processes.
        :param function: function name
        :param raster_path_list: input path
        :param use_value_as_nodata: list of integer values as nodata in
            calculation
        :param progress_message: progress message
        :param min_progress: minimum progress value
        :param max_progress: maximum progress value
        """
        cfg.logger.log.debug('multiprocess: %s' % function)
        self.output = False
        process_result = {}
        process_output_files = {}
        # calculation data type
        if calculation_datatype is None:
            calculation_datatype = [np.float32] * len(raster_path_list)
        if output_nodata_value is None:
            output_nodata_value = [cfg.nodata_val] * len(raster_path_list)
        if output_data_type is None:
            output_data_type = [cfg.raster_data_type] * len(raster_path_list)
        if n_processes is None:
            n_processes = self.n_processes
        elif n_processes > self.n_processes:
            self.start(n_processes)
        if n_processes > len(raster_path_list):
            n_processes = len(raster_path_list)
        if min_progress is None:
            min_progress = 0
        if max_progress is None:
            max_progress = 100
        # reserve progress for writing output
        max_progress = round(max_progress * 0.9)
        if max_progress < min_progress:
            max_progress = min_progress
        if progress_message is None:
            progress_message = 'processing'
        if available_ram is None:
            available_ram = cfg.available_ram
        # progress queue
        p_mq = self.manager.Queue()
        # calculate block size
        results = []
        # one process per raster
        if len(raster_path_list) <= n_processes:
            ranges = list(range(len(raster_path_list)))
        # n threads running 2 rasters and m running 1 raster
        # 2 * n + m = raster_number
        # n + m = n_processes
        elif len(raster_path_list) / n_processes < 2:
            n = len(raster_path_list) - n_processes
            ranges = list(range(0, n * 2, 2))
            ranges.extend(list(range(n * 2, len(raster_path_list))))
        # calculate rasters per process
        else:
            ranges = list(
                range(
                    0, len(raster_path_list),
                    int(len(raster_path_list) / n_processes)
                )
            )
        ranges.append(len(raster_path_list))
        for p in range(1, len(ranges)):
            raster_paths = raster_path_list[ranges[p - 1]: ranges[p]]
            if output_raster_list:
                output_list = output_raster_list[ranges[p - 1]: ranges[p]]
            else:
                output_list = [None]
            if output_band_number_list is None:
                output_band_number = [1] * len(raster_paths)
            else:
                output_band_number = output_band_number_list[
                                     ranges[p - 1]: ranges[p]]
            if scale:
                scl = scale[ranges[p - 1]: ranges[p]]
            else:
                scl = [None] * len(raster_paths)
            if offset:
                offs = offset[ranges[p - 1]: ranges[p]]
            else:
                offs = [None] * len(raster_paths)
            if use_value_as_nodata is None:
                use_value_as_nodata = [None] * len(raster_paths)
            process_parameters = [p, cfg.temp, available_ram / n_processes,
                                  cfg.gdal_path, p_mq, cfg.refresh_time]
            input_parameters = [raster_paths,
                                calculation_datatype[ranges[p - 1]: ranges[p]],
                                boundary_size, None, scl, offs,
                                use_value_as_nodata[ranges[p - 1]: ranges[p]],
                                None, input_nodata_as_value, multi_add_factors,
                                dummy_bands]
            output_parameters = [output_list,
                                 output_data_type[ranges[p - 1]: ranges[p]],
                                 compress, compress_format, any_nodata_mask,
                                 output_nodata_value[ranges[p - 1]: ranges[p]],
                                 output_band_number, keep_output_array,
                                 keep_output_argument]
            if function_argument:
                f_arg = function_argument[ranges[p - 1]: ranges[p]]
            else:
                f_arg = None
            if function_variable:
                f_var = function_variable[ranges[p - 1]: ranges[p]]
            else:
                f_var = None
            c = self.pool.apply_async(
                processor.function_initiator,
                args=(process_parameters, input_parameters, output_parameters,
                      function, f_arg, f_var, True, False, False)
            )
            results.append([c, p])
        while True:
            # update progress
            p_r = []
            for r in results:
                p_r.append(r[0].ready())
            if all(p_r):
                break
            time.sleep(cfg.refresh_time)
            # progress message
            try:
                p_m_qp = p_mq.get(False)
                count_progress = int(p_m_qp[0])
                length = int(p_m_qp[1])
                progress = int(100 * count_progress / length)
                cfg.progress.update(
                    message=progress_message, step=count_progress,
                    steps=length, minimum=min_progress,
                    maximum=max_progress, percentage=progress
                )
            except Exception as err:
                str(err)
        for r in results:
            res = r[0].get()
            process_result[r[1]] = res[0]
            process_output_files[r[1]] = res[1]
            cfg.logger.log.debug(res[3])
            # error
            if res[2] is not False:
                cfg.logger.log.error('multiprocess: %s' % res[2])
                messages.error('multiprocess: %s' % res[2])
                gc.collect()
                return
        gc.collect()
        multiprocess_dictionary = self._collect_results(
            process_result, process_output_files, None, scale=None,
            offset=None, output_nodata_value=None,
            output_data_type=output_data_type, compress=compress,
            compress_format=compress_format, n_processes=n_processes,
            available_ram=available_ram, min_progress=max_progress
        )
        cfg.progress.update(percentage=False)
        cfg.logger.log.debug('end')
        self.output = multiprocess_dictionary

    # collect multiprocess results
    def _collect_results(
            self, process_result, process_output_files, output_raster_path,
            scale=None, offset=None, virtual_raster=None,
            output_nodata_value=None, output_data_type=None, compress=False,
            compress_format='LZW', delete_array=None, n_processes=1,
            available_ram: int = None, min_progress=None, max_progress=None
    ):
        cfg.logger.log.debug('start')
        multiprocess_dictionary = {}
        # temporary raster output
        tmp_rast_list = []
        p = None
        # get parallel dictionary result
        for p in sorted(process_result):
            try:
                # following iteration
                multiprocess_dictionary[p].extend(process_result[p])
            except Exception as err:
                str(err)
                # first iteration
                multiprocess_dictionary[p] = process_result[p]
        # output files
        if output_raster_path is not None:
            try:
                len(process_output_files[p]) * len(process_output_files)
            except Exception as err:
                cfg.logger.log.error(str(err))
                gc.collect()
                messages.error(str(err))
                return False
            # collect temporary raster paths
            for r in process_output_files:
                for op in process_output_files[r]:
                    tmp_rast_list.append(op[0])
            # get unique rasters and sort by process name
            tmp_rast_list = sorted(list(set(tmp_rast_list)))
            cfg.logger.log.debug('tmp_rast_list: %s' % str(tmp_rast_list))
            if scale is not None:
                scl = scale
            else:
                scl = 1
            if offset is not None:
                offs = offset
            else:
                offs = 0
            if scale is not None or offset is not None:
                par_scale_offset = ' -a_scale %s -a_offset %s' % (
                    str(scl), str(offs))
            else:
                par_scale_offset = ''
            # virtual raster output
            if virtual_raster:
                tmp_list = []
                dir_path = files_directories.parent_directory(
                    output_raster_path
                )
                file_count = 1
                f_name = files_directories.file_name(output_raster_path, False)
                # move temporary files
                for tR in tmp_rast_list:
                    out_r = '%s/%s_%02d%s' % (
                        dir_path, f_name, file_count, cfg.tif_suffix)
                    files_directories.create_parent_directory(out_r)
                    file_count += 1
                    tmp_list.append(out_r)
                    files_directories.move_file(tR, out_r)
                # create virtual raster
                raster_vector.create_virtual_raster_2_mosaic(
                    input_raster_list=tmp_list, output=output_raster_path,
                    dst_nodata=output_nodata_value, relative_to_vrt=1,
                    data_type=output_data_type
                )
                # fix relative to vrt in xml
                raster_vector.force_relative_to_vrt(output_raster_path)
            # copy raster output
            else:
                vrt_file = cfg.temp.temporary_raster_path(
                    extension=cfg.vrt_suffix
                )
                try:
                    # create virtual raster
                    raster_vector.create_virtual_raster_2_mosaic(
                        input_raster_list=tmp_rast_list, output=vrt_file,
                        dst_nodata=output_nodata_value,
                        data_type=output_data_type
                    )
                    files_directories.create_parent_directory(
                        output_raster_path
                    )
                except Exception as err:
                    cfg.logger.log.error(str(err))
                    messages.error(str(err))
                try:
                    # copy raster
                    self.gdal_copy_raster(
                        vrt_file, output_raster_path, 'GTiff', compress,
                        compress_format, additional_params='-ot %s%s' % (
                            str(output_data_type), par_scale_offset),
                        n_processes=n_processes, min_progress=min_progress,
                        max_progress=max_progress
                    )
                except Exception as err:
                    cfg.logger.log.error(str(err))
                    messages.error(str(err))
                    try:
                        # try to create different virtual raster then copy
                        vrt_file = cfg.temp.temporary_raster_path(
                            extension=cfg.vrt_suffix
                        )
                        raster_vector.create_virtual_raster(
                            input_raster_list=tmp_rast_list, output=vrt_file,
                            nodata_value=output_nodata_value,
                            data_type=output_data_type
                        )
                        self.gdal_copy_raster(
                            vrt_file, output_raster_path, 'GTiff', compress,
                            compress_format, additional_params='-ot %s%s' % (
                                str(output_data_type), par_scale_offset),
                            available_ram=available_ram,
                            min_progress=min_progress,
                            max_progress=max_progress
                        )
                    except Exception as err:
                        cfg.logger.log.error(str(err))
                        gc.collect()
                        messages.error(str(err))
                        return False
        # delete temporary rasters
        if delete_array and not virtual_raster:
            for n in tmp_rast_list:
                try:
                    os.remove(n)
                except Exception as err:
                    str(err)
        gc.collect()
        cfg.logger.log.debug('end')
        return multiprocess_dictionary

    # collect classification results
    def _collect_classification_results(
            self, process_result, output_raster_path, function_argument=None,
            scale=None, offset=None, virtual_raster=None,
            output_nodata_value=None, output_data_type=None,
            compress=False, compress_format='LZW',
            delete_array=None, n_processes=1, available_ram: int = None,
            min_progress=None, max_progress=None
    ):
        cfg.logger.log.debug('start')
        # classification output
        output_classification_list = []
        # algorithm output
        output_algorithm_list = []
        # signatures output
        signature_raster_list = []
        # get parallel dictionary result
        for p in sorted(process_result):
            if len(output_classification_list) == 0:
                # first iteration
                output_classification_list = process_result[p][0][0]
            else:
                output_classification_list.extend(process_result[p][0][0])
            if len(output_algorithm_list) == 0:
                # first iteration
                output_algorithm_list = process_result[p][0][1]
            else:
                output_algorithm_list.extend(process_result[p][0][1])
            if len(signature_raster_list) == 0:
                # first iteration
                signature_raster_list = process_result[p][0][2]
            else:
                for s in process_result[p][0][2]:
                    if s in signature_raster_list:
                        signature_raster_list[s].extend(
                            process_result[p][0][2][s]
                        )
                    else:
                        signature_raster_list[s] = process_result[p][0][2][s]
        cfg.logger.log.debug(
            'output_classification_list: %s; output_algorithm_list: %s; '
            'signature_raster_list: %s'
            % (str(output_classification_list), str(output_algorithm_list),
               str(signature_raster_list))
        )
        # output files
        if output_raster_path is not None:
            signatures = function_argument[cfg.spectral_signatures_framework]
            # get selected signatures
            signatures_table = signatures.table[signatures.table.selected == 1]
            if scale is not None:
                scl = scale
            else:
                scl = 1
            if offset is not None:
                offs = offset
            else:
                offs = 0
            if scale is not None or offset is not None:
                par_scale_offset = ' -a_scale %s -a_offset %s' % (
                    str(scl), str(offs))
            else:
                par_scale_offset = ''
            # virtual raster output
            if virtual_raster:
                tmp_list = []
                dir_path = files_directories.parent_directory(
                    output_raster_path
                )
                f_count = 1
                f_name = files_directories.file_name(output_raster_path, False)
                # move temporary files
                for r in output_classification_list:
                    out_r = '%s/%s_%02d%s' % (
                        dir_path, f_name, f_count, cfg.tif_suffix)
                    files_directories.create_parent_directory(out_r)
                    f_count += 1
                    tmp_list.append(out_r)
                    files_directories.move_file(r, out_r)
                # create virtual raster
                raster_vector.create_virtual_raster_2_mosaic(
                    input_raster_list=tmp_list, output=output_raster_path,
                    dst_nodata=output_nodata_value, relative_to_vrt=1,
                    data_type=output_data_type
                )
                # fix relative to vrt in xml
                raster_vector.force_relative_to_vrt(output_raster_path)
                # algorithm raster
                if len(output_algorithm_list) > 0:
                    tmp_list = []
                    f_count = 1
                    # move temporary files
                    for r in output_algorithm_list:
                        out_r = '%s/%s_alg_%02d%s' % (
                            dir_path, f_name, f_count, cfg.tif_suffix)
                        f_count += 1
                        tmp_list.append(out_r)
                        files_directories.move_file(r, out_r)
                    # create virtual raster
                    raster_vector.create_virtual_raster_2_mosaic(
                        input_raster_list=tmp_list,
                        output='%s/%s_alg%s' % (
                            dir_path, f_name, cfg.vrt_suffix),
                        dst_nodata=output_nodata_value, relative_to_vrt=1,
                        data_type=output_data_type
                    )
                    # fix relative to vrt in xml
                    raster_vector.force_relative_to_vrt(
                        '%s/%s_alg%s' % (dir_path, f_name, cfg.vrt_suffix)
                    )
                # signature rasters
                if len(signature_raster_list) > 0:
                    for s in signature_raster_list:
                        tmp_list = []
                        sig_name = '%s_mc_%s_c_%s' % (
                            f_name, str(
                                signatures_table[
                                    signatures_table.signature_id ==
                                    s].macroclass_id[
                                    0]
                            ),
                            str(
                                signatures_table[
                                    signatures_table.signature_id ==
                                    s].class_id[
                                    0]
                            ))
                        # move temporary files
                        for r in signature_raster_list[s]:
                            out_r = '%s/%s_%02d%s' % (
                                dir_path, sig_name, f_count, cfg.tif_suffix)
                            f_count += 1
                            tmp_list.append(out_r)
                            files_directories.move_file(r, out_r)
                        # create virtual raster
                        raster_vector.create_virtual_raster_2_mosaic(
                            input_raster_list=tmp_list, output='%s/%s%s' % (
                                dir_path, sig_name, cfg.vrt_suffix),
                            dst_nodata=output_nodata_value, relative_to_vrt=1,
                            data_type=output_data_type
                        )
                        # fix relative to vrt in xml
                        raster_vector.force_relative_to_vrt(
                            '%s/%s%s' % (dir_path, sig_name, cfg.vrt_suffix)
                        )
            # copy raster output
            else:
                dir_path = files_directories.parent_directory(
                    output_raster_path
                )
                f_name = files_directories.file_name(output_raster_path, False)
                vrt_file = cfg.temp.temporary_raster_path(
                    extension=cfg.vrt_suffix
                )
                try:
                    # create virtual raster
                    raster_vector.create_virtual_raster_2_mosaic(
                        input_raster_list=output_classification_list,
                        output=vrt_file, dst_nodata=output_nodata_value,
                        data_type=output_data_type
                    )
                    files_directories.create_parent_directory(
                        output_raster_path
                    )
                except Exception as err:
                    cfg.logger.log.error(str(err))
                    messages.error(str(err))
                try:
                    # copy raster
                    self.gdal_copy_raster(
                        vrt_file, output_raster_path, 'GTiff', compress,
                        compress_format, additional_params='-ot %s%s' % (
                            str(output_data_type), par_scale_offset),
                        n_processes=n_processes, min_progress=min_progress,
                        max_progress=max_progress
                    )
                except Exception as err:
                    cfg.logger.log.error(str(err))
                    messages.error(str(err))
                    try:
                        # try to create different virtual raster then copy
                        vrt_file = cfg.temp.temporary_raster_path(
                            extension=cfg.vrt_suffix
                        )
                        raster_vector.create_virtual_raster(
                            input_raster_list=output_classification_list,
                            output=vrt_file, nodata_value=output_nodata_value,
                            data_type=output_data_type
                        )
                        self.gdal_copy_raster(
                            vrt_file, output_raster_path, 'GTiff', compress,
                            compress_format, additional_params='-ot %s%s' % (
                                str(output_data_type), par_scale_offset),
                            available_ram=available_ram,
                            min_progress=min_progress,
                            max_progress=max_progress
                        )
                    except Exception as err:
                        cfg.logger.log.error(str(err))
                        gc.collect()
                        messages.error(str(err))
                        return False
                # algorithm raster
                if len(output_algorithm_list) > 0:
                    vrt_file = cfg.temp.temporary_raster_path(
                        extension=cfg.vrt_suffix
                    )
                    try:
                        # create virtual raster
                        raster_vector.create_virtual_raster_2_mosaic(
                            input_raster_list=output_algorithm_list,
                            output=vrt_file, dst_nodata=output_nodata_value,
                            data_type=output_data_type
                        )
                    except Exception as err:
                        cfg.logger.log.error(str(err))
                        messages.error(str(err))
                    try:
                        # copy raster
                        self.gdal_copy_raster(
                            vrt_file,
                            '%s/%s_alg%s' % (dir_path, f_name, cfg.tif_suffix),
                            'GTiff', compress, compress_format,
                            additional_params='-ot %s%s' % (
                                str(output_data_type), par_scale_offset),
                            n_processes=n_processes, min_progress=min_progress,
                            max_progress=max_progress
                        )
                    except Exception as err:
                        cfg.logger.log.error(str(err))
                        messages.error(str(err))
                # signature rasters
                if len(signature_raster_list) > 0:
                    for s in signature_raster_list:
                        sig_name = '%s_mc_%s_c_%s' % (
                            f_name, str(
                                signatures_table[
                                    signatures_table.signature_id ==
                                    s].macroclass_id[
                                    0]
                            ),
                            str(
                                signatures_table[
                                    signatures_table.signature_id ==
                                    s].class_id[
                                    0]
                            ))
                        vrt_file = cfg.temp.temporary_raster_path(
                            extension=cfg.vrt_suffix
                        )
                        try:
                            # create virtual raster
                            raster_vector.create_virtual_raster_2_mosaic(
                                input_raster_list=signature_raster_list[s],
                                output=vrt_file,
                                dst_nodata=output_nodata_value,
                                data_type=output_data_type
                            )
                        except Exception as err:
                            cfg.logger.log.error(str(err))
                            messages.error(str(err))
                        try:
                            # copy raster
                            self.gdal_copy_raster(
                                vrt_file, '%s/%s%s' % (
                                    dir_path, sig_name, cfg.tif_suffix),
                                'GTiff', compress, compress_format,
                                additional_params='-ot %s%s' % (
                                    str(output_data_type), par_scale_offset),
                                n_processes=n_processes,
                                min_progress=min_progress,
                                max_progress=max_progress
                            )
                        except Exception as err:
                            cfg.logger.log.error(str(err))
                            messages.error(str(err))
        # delete temporary rasters
        if delete_array and not virtual_raster:
            output_classification_list.extend(output_algorithm_list)
            output_classification_list.extend(signature_raster_list)
            for n in output_classification_list:
                try:
                    os.remove(n)
                except Exception as err:
                    str(err)
        gc.collect()
        cfg.logger.log.debug('end')
        return output_raster_path

    # warp with GDAL
    def _gdal_warping(
            self, input_raster, output, output_format='GTiff', s_srs=None,
            t_srs=None, resample_method=None, raster_data_type=None,
            compression=None, compress_format='DEFLATE', additional_params='',
            n_processes: int = None, available_ram: int = None,
            src_nodata=None, dst_nodata=None
    ):
        cfg.logger.log.debug('start')
        out_dir = files_directories.parent_directory(output)
        files_directories.create_directory(out_dir)
        if resample_method is None:
            resample_method = 'near'
        elif resample_method == 'sum':
            gdal_v = raster_vector.get_gdal_version()
            if float('%s.%s' % (gdal_v[0], gdal_v[1])) < 3.1:
                cfg.logger.log.error('Error GDAL version')
                return False
        if n_processes is None:
            n_processes = cfg.n_processes
        op = ' -r %s -co BIGTIFF=YES -multi -wo NUM_THREADS=%s' % (
            resample_method, str(n_processes))
        if compression is None:
            if cfg.raster_compression:
                op += ' -co COMPRESS=%s' % compress_format
        elif compression:
            op += ' -co COMPRESS=%s' % compress_format
        if s_srs is not None:
            op += ' -s_srs %s' % s_srs
        if t_srs is not None:
            op += ' -t_srs %s' % t_srs
        if raster_data_type is not None:
            op += ' -ot %s' % raster_data_type
        if src_nodata is not None:
            op += ' -srcnodata %s' % str(src_nodata)
        if dst_nodata is not None:
            op += ' -dstnodata %s' % str(dst_nodata)
        op += ' -of %s' % output_format
        if additional_params is not None:
            op = ' %s %s' % (additional_params, op)
        p = 0
        # progress queue
        p_mq = self.manager.Queue()
        if available_ram is None:
            available_ram = cfg.available_ram
        available_ram = str(int(available_ram) * 1000000)
        process_parameters = [p, cfg.temp, available_ram, cfg.gdal_path, p_mq]
        cfg.logger.log.debug(
            'process_parameters: %s' % str(process_parameters)
        )
        results = []
        c = self.pool.apply_async(
            processor.gdal_warp,
            args=(input_raster, output, op, process_parameters)
        )
        results.append([c, p])
        while True:
            p_r = []
            for r in results:
                p_r.append(r[0].ready())
            if all(p_r):
                break
            time.sleep(cfg.refresh_time)
            # progress message
            try:
                p_m_qp = p_mq.get(False)
                progress = int(p_m_qp)
                cfg.progress.update(
                    message='writing raster', percentage=progress
                )
                cfg.logger.log.debug('progress: %s' % str(progress))
            except Exception as err:
                str(err)
        for r in results:
            res = r[0].get()
            cfg.logger.log.debug('res[3]: %s' % str(res[3]))
            # error
            if res[2] is not False:
                cfg.logger.log.error(
                    'Error proc %s-%s' % (str(p), str(res[1]))
                )
                return False
        cfg.logger.log.debug('end; output: %s' % str(output))
        return output

    # create warped virtual raster
    def create_warped_vrt(
            self, raster_path, output_path, output_wkt=None,
            align_raster_path=None, same_extent=False,
            n_processes: int = None, src_nodata=None, dst_nodata=None
    ):
        cfg.logger.log.debug('start')
        # additional params
        extra = None
        # calculate minimal extent
        if align_raster_path is not None:
            # align raster extent and pixel size
            try:
                (left_align, top_align, right_align, bottom_align, p_x_align,
                 p_y_align, output_wkt,
                 unit) = raster_vector.image_geotransformation(
                    align_raster_path
                )
                # check projections
                align_sys_ref = raster_vector.get_spatial_reference(output_wkt)
            except Exception as err:
                cfg.logger.log.error(str(err))
                return False
            # input_path raster extent and pixel size
            try:
                (left_input, top_input, right_input, bottom_input, p_x_input,
                 p_y_input, proj_input,
                 unit_input) = raster_vector.image_geotransformation(
                    raster_path
                )
                input_sys_ref = raster_vector.get_spatial_reference(proj_input)
                left_projected, top_projected = \
                    raster_vector.project_point_coordinates(
                        left_input, top_input, input_sys_ref, align_sys_ref
                    )
                right_projected, bottom_projected = \
                    raster_vector.project_point_coordinates(
                        right_input, bottom_input, input_sys_ref, align_sys_ref
                    )
            # Error latitude or longitude exceeded limits
            except Exception as err:
                cfg.logger.log.error(str(err))
                return False
            if not same_extent:
                # minimum extent
                if left_projected < left_align:
                    left_r = left_align - int(
                        2 + (left_align - left_projected) / p_x_align
                    ) * p_x_align
                else:
                    left_r = left_align + int(
                        (left_projected - left_align) / p_x_align - 2
                    ) * p_x_align
                if right_projected > right_align:
                    right_r = right_align + int(
                        2 + (right_projected - right_align) / p_x_align
                    ) * p_x_align
                else:
                    right_r = right_align - int(
                        (right_align - right_projected) / p_x_align - 2
                    ) * p_x_align
                if top_projected > top_align:
                    top_r = top_align + int(
                        2 + (top_projected - top_align) / p_y_align
                    ) * p_y_align
                else:
                    top_r = top_align - int(
                        (top_align - top_projected) / p_y_align - 2
                    ) * p_y_align
                if bottom_projected > bottom_align:
                    bottom_r = bottom_align + int(
                        (bottom_projected - bottom_align) / p_y_align - 2
                    ) * p_y_align
                else:
                    bottom_r = bottom_align - int(
                        2 + (bottom_align - bottom_projected) / p_y_align
                    ) * p_y_align
            else:
                left_r = left_align
                right_r = right_align
                top_r = top_align
                bottom_r = bottom_align
            extra = '-tr %s %s -te %s %s %s %s' % (
                str(p_x_align), str(p_y_align), str(left_r), str(bottom_r),
                str(right_r), str(top_r))
        self._gdal_warping(
            input_raster=raster_path, output=output_path, output_format='VRT',
            s_srs=None, t_srs=output_wkt, additional_params=extra,
            n_processes=n_processes, src_nodata=src_nodata,
            dst_nodata=dst_nodata
        )
        # workaround to gdalwarp issue ignoring scale and offset
        try:
            (gt, r_p, unit, xy_count, nd, number_of_bands, block_size,
             scale_offset, data_type) = raster_vector.raster_info(raster_path)
            scale = scale_offset[0]
            offset = scale_offset[1]
            if scale != 1 or offset != 0:
                o_r = raster_vector.open_raster(output_path)
                b_o = o_r.GetRasterBand(1)
                try:
                    b_o.SetScale(scale)
                except Exception as err:
                    str(err)
                try:
                    b_o.SetOffset(offset)
                except Exception as err:
                    str(err)
        except Exception as err:
            cfg.logger.log.error(str(err))
        cfg.logger.log.debug('end; output_path: %s' % output_path)
        return output_path

    # run multiprocess join table
    def join_tables_multiprocess(
            self, table1, table2, field1_name, field2_name, nodata_value=None,
            join_type=None, postfix=None, n_processes: int = None,
            progress_message=None, min_progress=None, max_progress=None
    ):
        """
        :param table1: input numpy table 1
        :param table2: input numpy table 2
        :param field1_name: input field table 1
        :param field2_name: input field table 2
        :param nodata_value: input nodata value
        :param join_type: join type
        :param postfix: postfix string
        :param n_processes: number of parallel processes.
        :param progress_message: progress message
        :param min_progress: minimum progress value
        :param max_progress: maximum progress value
        """
        cfg.logger.log.debug('multiprocess join')
        # progress queue
        p_mq = self.manager.Queue()
        self.output = False
        process_result = {}
        if n_processes is None:
            n_processes = self.n_processes
        elif n_processes > self.n_processes:
            self.start(n_processes)
        if min_progress is None:
            min_progress = 0
        if max_progress is None:
            max_progress = 100
        if progress_message is None:
            progress_message = 'processing'
        # output field names
        output_names = list(table1.dtype.names)
        # output field dtypes
        output_dtypes = []
        # fields table 1
        for name in table1.dtype.names:
            output_dtypes.append(table1[name].dtype)
        # table 1 field names
        table1_names = list(table1.dtype.names)
        # table 1 field dtypes
        table1_dtypes = []
        for name in table1.dtype.names:
            table1_dtypes.append(table1[name].dtype)
        # table 2 field names and dtypes
        table2_dtypes = []
        table2_names = []
        table2_output_names = []
        for name in table2.dtype.names:
            if field2_name != str(name):
                table2_names.append(str(name))
                # rename duplicate field names
                if str(name) not in table1.dtype.names:
                    output_names.append(str(name))
                    table2_output_names.append(str(name))
                else:
                    output_names.append('%s%s' % (str(name), postfix))
                    table2_output_names.append('%s%s' % (str(name), postfix))
                output_dtypes.append(table2[name].dtype)
                table2_dtypes.append(table2[name].dtype)
        # identify table features
        table1_features = table1_features_index = table2_features_index = None
        features_table_2_outer = None
        if join_type == 'left':
            # unique features
            table1_features = list(np.unique(table1[field1_name]))
            table_2_unique = np.unique(table2[field2_name], return_index=True)
            table2_features = table_2_unique[0]
            table2_features_index = table_2_unique[1]
            # main features
            features = table1_features
        elif join_type == 'right':
            # unique features
            table_1_unique = np.unique(table1[field1_name], return_index=True)
            table1_features = table_1_unique[0]
            table1_features_index = table_1_unique[1]
            table2_features = list(np.unique(table2[field2_name]))
            # main features
            features = table2_features
        elif join_type == 'inner':
            # find common features
            table1_unique = np.unique(table1[field1_name])
            table_2_unique = np.unique(table2[field2_name], return_index=True)
            table2_features = table_2_unique[0]
            table2_features_index = table_2_unique[1]
            features = list(np.intersect1d(table1_unique, table2_features))
        elif join_type == 'outer':
            # find all unique features
            table1_features = list(np.unique(table1[field1_name]))
            table_2_unique = np.unique(table2[field2_name], return_index=True)
            table2_features = table_2_unique[0]
            table2_features_index = table_2_unique[1]
            # main features as left join
            features = table1_features
            # features from table 2 not in table 1
            features_table_2_outer = list(
                np.setdiff1d(table2_features, table1_features)
            )
        else:
            self.output = False
            cfg.logger.log.error('join type not found')
            return
        features_count = len(features)
        # calculate process ranges
        if features_count < n_processes:
            n_processes = features_count
        try:
            sections_range = list(
                range(0, features_count, round(features_count / n_processes))
            )
        except Exception as err:
            str(err)
            sections_range = [0]
        sections_range.append(features_count)
        cfg.logger.log.debug(
            'features_count: %s; len(sections_range): %s'
            % (str(features_count), str(len(sections_range)))
        )
        ranges = []
        for index_range in range(1, len(sections_range)):
            ranges.append(
                features[sections_range[index_range - 1]:sections_range[
                    index_range]]
            )
        results = []
        for p in range(len(ranges)):
            feature_range = ranges[p]
            table_1_parameters = [table1, field1_name, table1_names,
                                  table1_dtypes, table1_features,
                                  table1_features_index]
            table_2_parameters = [table2, field2_name, table2_names,
                                  table2_dtypes, table2_features,
                                  table2_features_index, table2_output_names,
                                  features_table_2_outer]
            process_parameters = [p, cfg.temp, p_mq, cfg.refresh_time]
            c = self.pool.apply_async(
                processor.table_join,
                args=(table_1_parameters, table_2_parameters, nodata_value,
                      join_type, feature_range, output_names,
                      process_parameters)
            )
            results.append([c, p])
        while True:
            # update progress
            p_r = []
            for r in results:
                p_r.append(r[0].ready())
            if all(p_r):
                break
            time.sleep(cfg.refresh_time)
            # progress message
            try:
                p_m_qp = p_mq.get(False)
                count_progress = int(p_m_qp[0])
                length = int(p_m_qp[1])
                progress = int(100 * count_progress / length)
                cfg.progress.update(
                    message=progress_message, step=count_progress,
                    steps=length, minimum=min_progress,
                    maximum=max_progress, percentage=progress
                )
            except Exception as err:
                str(err)
        for r in results:
            res = r[0].get()
            process_result[r[1]] = res[0]
            cfg.logger.log.debug(res[3])
            # error
            if res[2] is not False:
                cfg.logger.log.error('multiprocess: %s' % str(res[2]))
                messages.error('multiprocess: %s' % str(res[2]))
                gc.collect()
                return
        gc.collect()
        output_table = _join_results(process_result)
        cfg.progress.update(percentage=False)
        cfg.logger.log.debug('end')
        self.output = output_table

    # convert raster to vector
    def multiprocess_raster_to_vector(
            self, raster_path, output_vector_path, field_name=None,
            n_processes: int = None,
            dissolve_output=True, min_progress=0, max_progress=100,
            available_ram: int = None
    ):
        max_progress_1 = int(max_progress / 3)
        cfg.logger.log.debug('start')
        process_result = []
        self.output = False
        if available_ram is None:
            available_ram = cfg.available_ram
        # raster blocks
        memory_unit = cfg.memory_unit_array_8
        (raster_x_size, raster_y_size, block_size_x, block_size_y,
         list_range_x, list_range_y, tot_blocks, number_of_bands) = \
            _calculate_block_size(
                raster_path, n_processes, memory_unit, 0,
                available_ram=available_ram
            )
        results = []
        # temporary raster output
        tmp_rast_list = []
        # calculate raster ranges
        for x in list_range_x:
            bs_x = block_size_x
            if x + bs_x > raster_x_size:
                bs_x = raster_x_size - x
            for y in list_range_y:
                bs_y = block_size_y
                if y + bs_y > raster_y_size:
                    bs_y = raster_y_size - y
                vrt_file = cfg.temp.temporary_raster_path(
                    extension=cfg.vrt_suffix
                )
                tmp_rast_list.append(vrt_file)
                # create virtual raster
                raster_vector.create_virtual_raster(
                    input_raster_list=[raster_path], output=vrt_file,
                    relative_extent_list=[x, y, bs_x, bs_y]
                )
        if field_name is None:
            field_name = 'DN'
        # progress queue
        p_mq = self.manager.Queue()
        # multiple parallel processes
        for p in range(len(tmp_rast_list)):
            vrt_path = tmp_rast_list[p]
            process_parameters = [p, cfg.temp, cfg.gdal_path, p_mq, int(
                int(available_ram) * 1000000 / len(tmp_rast_list)
            )]
            t_vector = cfg.temp.temporary_raster_path(
                extension=cfg.gpkg_suffix
            )
            c = self.pool.apply_async(
                processor.raster_to_vector_process,
                args=(vrt_path, t_vector, field_name, process_parameters)
            )
            results.append([c, p])
        while True:
            # update progress
            p_r = []
            for r in results:
                p_r.append(r[0].ready())
            if all(p_r):
                break
            time.sleep(cfg.refresh_time)
            # progress message
            try:
                progress = round(p_mq.get(False))
                step = round(
                    min_progress + progress * (
                            max_progress_1 - min_progress) / 100
                )
                cfg.progress.update(
                    message='processing to vector', step=step,
                    percentage=progress
                )
            except Exception as err:
                str(err)
        # get results
        for r in results:
            res = r[0].get()
            process_result.append(res[0])
            cfg.logger.log.debug(res[2])
            # error
            if res[1] is not False:
                cfg.logger.log.error('error multiprocess: %s' % str(res[1]))
                gc.collect()
                return
        gc.collect()
        cfg.progress.update(
            message='merging vectors', step=max_progress_1, percentage=False
        )
        # merge layers to new layer
        if len(tmp_rast_list) > 1:
            # input layers
            input_layers_list = []
            # y coordinates of polygons on borders
            y_list = []
            for i in range(len(process_result)):
                vector_path, min_y, max_y = process_result[i]
                input_layers_list.append(vector_path)
                if i > 0:
                    if max_y is not None:
                        y_list.append(str(max_y))
                if i < len(tmp_rast_list) - 1:
                    if min_y is not None:
                        y_list.append(str(min_y))
            if dissolve_output:
                # merged vector
                t_vector = cfg.temp.temporary_raster_path(
                    extension=cfg.gpkg_suffix
                )
            else:
                t_vector = output_vector_path
            merge = raster_vector.merge_all_layers(
                input_layers_list, t_vector, min_progress=max_progress_1,
                max_progress=max_progress_1 * 2,
                dissolve_output=dissolve_output
            )
            if dissolve_output:
                merge = raster_vector.merge_dissolve_layer(
                    t_vector, output_vector_path, field_name, y_list,
                    min_progress=max_progress_1 * 2, max_progress=max_progress
                )
        else:
            merge = tmp_rast_list[0]
        self.output = merge
        cfg.logger.log.debug('end')

    # convert raster sieve
    def multiprocess_raster_sieve(
            self, raster_path, n_processes: int = None, sieve_size=None,
            connected=None, output_nodata_value=None, output=None,
            output_data_type=None, compress=None, compress_format=None,
            available_ram: int = None, min_progress=0, max_progress=100
    ):
        cfg.logger.log.debug('start')
        if compress_format is None:
            compress_format = 'LZW'
        # progress queue
        p_mq = self.manager.Queue()
        results = []
        self.output = False
        if available_ram is None:
            available_ram = cfg.available_ram
        available_ram = int(available_ram / (n_processes * 2))
        # temporary raster output
        process_result = {}
        process_output_files = {}
        p = 0
        process_parameters = [p, cfg.temp, cfg.gdal_path, p_mq, available_ram]
        input_parameters = [raster_path, sieve_size, connected]
        output_parameters = [output, output_data_type, compress,
                             compress_format, output_nodata_value]
        c = self.pool.apply_async(
            processor.raster_sieve_process,
            args=(process_parameters, input_parameters,
                  output_parameters)
        )
        results.append([c, p])
        while True:
            # update progress
            p_r = []
            for r in results:
                p_r.append(r[0].ready())
            if all(p_r):
                break
            time.sleep(cfg.refresh_time)
            # progress message
            try:
                p_m_qp = p_mq.get(False)
                progress = int(p_m_qp)
                cfg.progress.update(
                    step=progress, steps=100, minimum=min_progress,
                    maximum=max_progress, percentage=progress
                )
            except Exception as err:
                str(err)
        for r in results:
            res = r[0].get()
            process_result[r[1]] = res[0]
            process_output_files[r[1]] = res[0]
            cfg.logger.log.debug(res[2])
            # error
            if res[1] is not False:
                cfg.logger.log.error('error multiprocess: %s' % str(res[1]))
                gc.collect()
                return
        gc.collect()
        cfg.progress.update(percentage=False)
        self.output = output
        cfg.logger.log.debug('end')

    # sum the output of multiprocess
    def multiprocess_sum_array(self, nodata=None):
        cfg.logger.log.debug('start')
        multiprocess_dictionary: Union[dict, bool] = self.output
        if not multiprocess_dictionary:
            cfg.logger.log.error('unable to process')
            return
        # calculate unique values and sum
        values = np.array([])
        sum_val = np.array([])
        for x in sorted(multiprocess_dictionary):
            for arr_x in multiprocess_dictionary[x]:
                try:
                    values = np.append(values, arr_x[1][0, ::].ravel())
                    sum_val = np.append(sum_val, arr_x[1][1, ::].ravel())
                except Exception as err:
                    cfg.logger.log.error(str(err))
                    return
        unique = list(np.unique(values, return_counts=False))
        val_sum = []
        for v in unique:
            if v != nodata:
                val_sum.append([v, sum_val[values == v].sum()])
        # dictionary of values and sum
        dtype_list = [('new_val', 'int64'), ('sum', 'int64')]
        unique_val = np.rec.fromarrays(np.asarray(val_sum).T, dtype=dtype_list)
        cfg.logger.log.debug(
            'end; unique_val.shape: %s'
            % str(unique_val.shape)
            )
        self.output = unique_val

    # get dictionary of sums
    def get_dictionary_sum(self):
        cfg.logger.log.debug('start')
        multiprocess_dictionary: Union[dict, bool] = self.output
        if not multiprocess_dictionary:
            cfg.logger.log.error('unable to process')
            return
        _dict = {}
        # get dictionaries
        for x in sorted(multiprocess_dictionary):
            try:
                for ar in multiprocess_dictionary[x]:
                    for i in ar[1]:
                        # sum values if multiple processes
                        if i in _dict:
                            _dict[i] = _dict[i] + ar[1][i]
                        else:
                            _dict[i] = ar[1][i]
            except Exception as err:
                cfg.logger.log.error(str(err))
                return
        cfg.logger.log.debug('end; _dict: %s' % str(_dict))
        self.output = _dict

    # find minimum DN
    def find_minimum_dn(self):
        cfg.logger.log.debug('start')
        multiprocess_dictionary: Union[dict, bool] = self.output
        if not multiprocess_dictionary:
            cfg.logger.log.error('unable to process')
            return
        dn_minimum_list = []
        for x in sorted(multiprocess_dictionary):
            try:
                # each array is an input
                for ar in multiprocess_dictionary[x]:
                    values = ar[1][0, ::].tolist()
                    count = ar[1][1, ::]
                    count_list = count.tolist()
                    total = count.sum()
                    min_threshold = total * 0.0001
                    sum_v = 0
                    i = 0
                    for v in values:
                        sum_v = sum_v + count_list[i]
                        i += 1
                        if sum_v >= min_threshold:
                            dn_minimum_list.append(v)
                            break
            except Exception as err:
                cfg.logger.log.error(str(err))
                return
        cfg.logger.log.debug('end; dn_minimum_list: %s' % str(dn_minimum_list))
        self.output = dn_minimum_list

    # roi arrays from output of multiprocess
    def multiprocess_roi_arrays(self):
        cfg.logger.log.debug('start')
        multiprocess_dictionary: Union[dict, bool] = self.output
        if not multiprocess_dictionary:
            cfg.logger.log.error('unable to process')
            return
        array_values = {}
        # calculate values
        for x in sorted(multiprocess_dictionary):
            for arr_x in multiprocess_dictionary[x]:
                for s in arr_x[1]:
                    try:
                        if s in array_values:
                            array_values[s].append(arr_x[1][s])
                        else:
                            array_values[s] = [arr_x[1][s]]
                    except Exception as err:
                        cfg.logger.log.error(str(err))
                        return
        self.output = array_values
        cfg.logger.log.debug('end')

    # spectral signature from output of multiprocess
    def multiprocess_spectral_signature(self):
        cfg.logger.log.debug('start')
        multiprocess_dictionary: Union[dict, bool] = self.output
        if not multiprocess_dictionary:
            cfg.logger.log.error('unable to process')
            return
        mean_values = []
        std = []
        # calculate values
        for x in sorted(multiprocess_dictionary):
            for arr_x in multiprocess_dictionary[x]:
                try:
                    mean_values.append(arr_x[1][0])
                    std.append(arr_x[1][1])
                except Exception as err:
                    cfg.logger.log.error(str(err))
                    return
        self.output = mean_values, std
        cfg.logger.log.debug('end')

    # unique values from output of multiprocess
    def multiprocess_unique_values(self):
        cfg.logger.log.debug('start')
        multiprocess_dictionary: Union[dict, bool] = self.output
        if not multiprocess_dictionary:
            cfg.logger.log.error('unable to process')
            return
        # calculate unique values
        values = None
        for x in sorted(multiprocess_dictionary):
            for arr_x in multiprocess_dictionary[x]:
                try:
                    if values is None:
                        values = arr_x[1]
                    else:
                        try:
                            values = np.vstack((values, arr_x[1]))
                        except Exception as err:
                            cfg.logger.log.error(str(err))
                except Exception as err:
                    cfg.logger.log.error(str(err))
                    return
        cfg.logger.log.debug('len(values): %s' % str(len(values)))
        # adapted from Jaime answer at
        # https://stackoverflow.com/questions/16970982/find-unique-rows-in
        # -numpy-array
        try:
            b_values = values.view(
                np.dtype((np.void, values.dtype.itemsize * values.shape[1]))
            )
            ff, index_a = np.unique(
                b_values, return_index=True, return_counts=False
            )
            output = values[index_a].tolist()
            self.output = output
            cfg.logger.log.debug('end; len(output): %s' % str(len(output)))
        except Exception as err:
            cfg.logger.log.error(str(err))
            return

    # run scikit process
    def run_scikit(
            self, function, classifier_list=None,
            list_train_argument_dictionaries=None, n_processes=None,
            available_ram: int = None, min_progress=None, max_progress=None
    ):
        cfg.logger.log.debug('start')
        self.output = False
        if n_processes is None:
            n_processes = self.n_processes
        elif n_processes > self.n_processes:
            self.start(n_processes)
        if available_ram is None:
            available_ram = cfg.available_ram
        if min_progress is None:
            min_progress = 0
        if max_progress is None:
            max_progress = 100
        # divide argument dictionaries for every process
        len_list_of_argument_dictionaries = len(
            list_train_argument_dictionaries
        )
        if len_list_of_argument_dictionaries <= n_processes:
            ranges = list(range(len_list_of_argument_dictionaries))
        # n threads running 2 dictionaries and m running 1 dictionary
        # 2 * n + m = len_list_of_argument_dictionaries
        # n + m = n_processes
        elif len_list_of_argument_dictionaries / n_processes < 2:
            n = len_list_of_argument_dictionaries - n_processes
            ranges = list(range(0, n * 2, 2))
            ranges.extend(
                list(range(n * 2, len_list_of_argument_dictionaries))
            )
        # calculate dictionary per process
        else:
            ranges = list(
                range(
                    0, len_list_of_argument_dictionaries,
                    int(len_list_of_argument_dictionaries / n_processes)
                )
            )
        ranges.append(len_list_of_argument_dictionaries)
        argument_dict_list = []
        list_of_classifier_process = []
        len_of_argument_dict_list = []
        for i in range(1, len(ranges)):
            argument_dict_list.append(
                list_train_argument_dictionaries[ranges[i - 1]: ranges[i]]
            )
            list_of_classifier_process.append(
                classifier_list[ranges[i - 1]: ranges[i]]
            )
            len_of_argument_dict_list.append(
                len(list_train_argument_dictionaries[ranges[i - 1]: ranges[i]])
            )
        logger_argument_index = len_of_argument_dict_list.index(
            max(len_of_argument_dict_list)
        )
        # result list
        results = []
        # iterate over arg dict list
        for p in range(0, len(argument_dict_list)):
            cfg.logger.log.debug('process %s' % str(p))
            if p == logger_argument_index:
                log_process = True
            else:
                log_process = False
            process_parameters = [p, cfg.temp, available_ram, log_process,
                                  cfg.logger]
            c = self.pool.apply_async(
                function, args=(
                    process_parameters, list_of_classifier_process[p],
                    argument_dict_list[p])
            )
            results.append([c, p])
        # divide progress max for process length
        max_progress_part = int(max_progress / max(len_of_argument_dict_list))
        max_progress_process = max_progress_part
        old_progress = 0
        while True:
            p_r = []
            for r in results:
                p_r.append(r[0].ready())
            if all(p_r):
                break
            time.sleep(cfg.refresh_time)
            # progress message
            try:
                # read progress from file
                with open('%s/scikit' % cfg.temp.dir, 'r') as f:
                    progress_line = f.readlines()[-1].split(' ')
                    progress = round(
                        int(progress_line[-3]) / int(
                            progress_line[-1].replace('\\n', '')
                        ) * 100
                    )
                cfg.progress.update(
                    message='fitting', step=progress, steps=100,
                    minimum=min_progress,
                    maximum=max_progress_process, percentage=progress
                )
                # scale progress for next process
                if progress < old_progress:
                    min_progress = int(max_progress_process)
                    max_progress_process += max_progress_part
                old_progress = int(progress)
            except Exception as err:
                str(err)
        # get results
        process_result = []
        for r in results:
            res = r[0].get()
            process_result.append(res[0])
            cfg.logger.log.debug(res[2])
            # error
            if res[1] is not False:
                cfg.logger.log.error('error multiprocess: %s' % str(res[1]))
                gc.collect()
                return
        gc.collect()
        self.output = process_result
        cfg.logger.log.debug('end; function: %s' % str(function))

    # run iterative process
    def run_iterative_process(
            self, function_list, argument_list, min_progress=None,
            max_progress=None, n_processes=None
    ):
        cfg.logger.log.debug('start')
        self.output = False
        if n_processes is None:
            n_processes = self.n_processes
        elif n_processes > self.n_processes:
            self.start(n_processes)
        if min_progress is None:
            min_progress = 0
        if max_progress is None:
            max_progress = 100
        # progress queue
        p_mq = self.manager.Queue()
        # divide argument dictionaries for every process
        len_list_of_argument_dictionaries = len(argument_list)
        if len_list_of_argument_dictionaries <= n_processes:
            ranges = list(range(len_list_of_argument_dictionaries))
        # n threads running 2 dictionaries and m running 1 dictionary
        # 2 * n + m = len_list_of_argument_dictionaries
        # n + m = n_processes
        elif len_list_of_argument_dictionaries / n_processes < 2:
            n = len_list_of_argument_dictionaries - n_processes
            ranges = list(range(0, n * 2, 2))
            ranges.extend(
                list(range(n * 2, len_list_of_argument_dictionaries))
            )
        # calculate dictionary per process
        else:
            ranges = list(
                range(
                    0, len_list_of_argument_dictionaries,
                    int(len_list_of_argument_dictionaries / n_processes)
                )
            )
        ranges.append(len_list_of_argument_dictionaries)
        argument_dict_list = []
        for i in range(1, len(ranges)):
            argument_dict_list.append(argument_list[ranges[i - 1]: ranges[i]])
        # result list
        results = []
        # iterate over arg dict list
        for p in range(0, len(argument_dict_list)):
            cfg.logger.log.debug('process %s' % str(p))
            if p == 0:
                progress_queue = p_mq
            else:
                progress_queue = None
            c = self.pool.apply_async(
                function_list[p], args=(
                    p, progress_queue, argument_dict_list[p], cfg.logger)
            )
            results.append([c, p])
        while True:
            p_r = []
            for r in results:
                p_r.append(r[0].ready())
            if all(p_r):
                break
            time.sleep(cfg.refresh_time)
            # progress message
            try:
                p_m_qp = p_mq.get(False)
                count_progress = int(p_m_qp[0])
                length = int(p_m_qp[1])
                progress = int(100 * count_progress / length)
                cfg.progress.update(
                    message='processing', step=progress, steps=100,
                    minimum=min_progress,
                    maximum=max_progress, percentage=progress
                )
            except Exception as err:
                str(err)
        # get results
        process_result = []
        for r in results:
            res = r[0].get()
            process_result.append(res[0])
            cfg.logger.log.debug(res[2])
            # error
            if res[1] is not False:
                cfg.logger.log.error('error multiprocess: %s' % str(res[1]))
                gc.collect()
                return
        gc.collect()
        self.output = process_result
        cfg.logger.log.debug('end; function_list: %s' % str(function_list))

    # copy raster with GDAL
    def gdal_copy_raster(
            self, input_raster, output, output_format='GTiff', compress=False,
            compress_format='DEFLATE',
            additional_params='', n_processes=1, available_ram: int = None,
            min_progress=None, max_progress=None
    ):
        cfg.logger.log.debug('start')
        # progress queue
        p_mq = self.manager.Queue()
        out_dir = files_directories.parent_directory(output)
        files_directories.create_directory(out_dir)
        op = ' -co BIGTIFF=YES -co NUM_THREADS=%s' % str(n_processes)
        if not compress:
            op += ' -of %s' % output_format
        else:
            op += ' -co COMPRESS=%s -of %s' % (
                compress_format, output_format)
        parameters = '%s %s' % (additional_params, op)
        p = 0
        if available_ram is None:
            available_ram = cfg.available_ram
        available_ram = str(int(available_ram) * 1000000)
        process_parameters = [p, cfg.temp, available_ram, cfg.gdal_path,
                              p_mq]
        cfg.logger.log.debug(str(process_parameters))
        results = []
        c = self.pool.apply_async(
            processor.gdal_translate, args=(
                input_raster, output, parameters, process_parameters)
        )
        results.append([c, p])
        while True:
            p_r = []
            for r in results:
                p_r.append(r[0].ready())
            if all(p_r):
                break
            time.sleep(cfg.refresh_time)
            # progress message
            try:
                p_m_qp = p_mq.get(False)
                progress = round(p_m_qp)
                cfg.progress.update(
                    message='writing raster', step=progress, steps=100,
                    minimum=min_progress, maximum=max_progress,
                    percentage=progress
                )
            except Exception as err:
                str(err)
        for r in results:
            res = r[0].get()
            # log
            cfg.logger.log.debug(res[3])
            # error
            if res[2] is not False:
                cfg.logger.log.error(
                    'Error proc %s-%s' % (str(p), str(res[1]))
                )
                return False
        cfg.logger.log.debug('end; output: %s' % str(output))
        return output


# calculate block size and pixel ranges
def _calculate_block_size(
        raster_path, n_processes, memory_unit, dummy_bands,
        available_ram: int = None
):
    cfg.logger.log.debug('start')
    info = raster_vector.raster_info(raster_path)
    if info is not False:
        (gt, crs, crs_unit, xy_count, nd, number_of_bands, block_size,
         scale_offset, data_type) = info
    else:
        cfg.logger.log.error(
            'unable to get raster info: %s' % str(raster_path)
        )
        return False
    r_x, r_y = xy_count
    # list of range pixels
    y_block = block_size[1]
    if y_block == r_y:
        y_block = block_size[0]
    single_block_size = r_x * y_block * memory_unit * (
            number_of_bands + dummy_bands)
    if available_ram is None:
        available_ram = cfg.available_ram
    ram_blocks = int(available_ram / (single_block_size * n_processes))
    if ram_blocks == 0:
        ram_blocks = 1
    block_size_x = r_x
    block_size_y = ram_blocks * y_block
    if block_size_x > r_x:
        block_size_x = r_x
    list_range_x = list(range(0, r_x, block_size_x))
    if block_size_y > r_y:
        block_size_y = int(r_y / float(n_processes)) + 1
    list_range_y = list(range(0, r_y, block_size_y))
    if len(list_range_y) < n_processes:
        block_size_y = int(r_y / float(n_processes)) + 1
        list_range_y = list(range(0, r_y, block_size_y))
    tot_blocks = len(list_range_x) * len(list_range_y)
    cfg.logger.log.debug(
        'end; r_x: %s; r_y: %s; block_size_x: %s; block_size_y: %s; '
        'list_range_x: %s; list_range_y: %s; tot_blocks: %s; '
        'number_of_bands: %s' % (
            r_x, r_y, block_size_x, block_size_y, list_range_x, list_range_y,
            tot_blocks, number_of_bands)
    )
    return (r_x, r_y, block_size_x, block_size_y, list_range_x, list_range_y,
            tot_blocks, number_of_bands)


# compute a raster piece (portions of raster to be processed) for each
# process divided in sections along the axis y
def _compute_raster_pieces(
        raster_x_size, raster_y_size, block_size_x, block_size_y, list_range_y,
        n_processes, separate_bands=False, boundary_size=None,
        unique_section=False
):
    """
    :param raster_x_size: number of raster columns
    :param raster_y_size: number of raster rows
    :param block_size_x: raster block size for columns
    :param block_size_y: raster block size for rows
    :param list_range_y: list of y pieces using block_size_y
    :param n_processes: number of parallel processes. to divide pieces
    :param separate_bands: if True, calculate a section for each raster piece
    :param boundary_size: pixel number to be added as boundary to each
    section along the axis y
    :param unique_section: if True, consider the whole raster as unique section
    """
    # list of pieces (one per process) where each range is based on sections
    pieces = []
    if unique_section:
        sections = [processor.RasterSection(
            x_min=0, y_min=0, x_max=raster_x_size, y_max=raster_y_size,
            x_size=raster_x_size, y_size=raster_y_size
        )]
        pieces.append(
            processor.RasterPiece(
                section_list=sections, x_min=0, y_min=0, x_max=raster_x_size,
                y_max=raster_y_size, x_size=raster_x_size, y_size=raster_y_size
            )
        )
    else:
        # create a list of indices of y pieces for each process
        try:
            y_range_index_list = list(
                range(
                    0, len(list_range_y),
                    round(len(list_range_y) / n_processes)
                )
            )
        except Exception as err:
            str(err)
            y_range_index_list = [0]
        y_range_index_list.append(len(list_range_y))
        # iterate over range index list
        for y_range_index in range(1, len(y_range_index_list)):
            # range variables
            y_min_range = list_range_y[y_range_index_list[y_range_index - 1]]
            y_max_range = None
            y_size_range = 0
            y_size_no_boundary_range = 0
            y_min_no_boundary_range_list = []
            y_max_no_boundary_range_list = []
            y_size_boundary_top_range = None
            y_size_boundary_bottom_range = None
            # list of Sections per process
            sections_list = []
            # iterate over process y pieces
            for process_y_range in range(
                    y_range_index_list[y_range_index - 1],
                    y_range_index_list[y_range_index]
            ):
                y = list_range_y[process_y_range]
                # multiple band processes
                if not separate_bands:
                    # without boundary
                    if boundary_size is None:
                        # section y size
                        y_size = block_size_y
                        # adapt to raster y size
                        if y + y_size > raster_y_size:
                            y_size = raster_y_size - y
                        y_max = y + y_size
                        sections_list.append(
                            processor.RasterSection(
                                x_min=0, y_min=y, x_max=block_size_x,
                                y_max=y_max,
                                x_size=block_size_x, y_size=y_size
                            )
                        )
                        # range variables
                        y_size_range += y_size
                        y_size_no_boundary_range = y_size_range
                        y_min_no_boundary_range_list.append(y)
                        y_max_no_boundary_range_list.append(y_max)
                    # with y boundary
                    else:
                        y_min_no_boundary = y
                        # nominal boundary
                        y_size_boundary_top = boundary_size
                        y_size_boundary_bottom = boundary_size
                        # nominal section y size without boundary
                        y_size_no_boundary = block_size_y
                        # minimum y with boundary
                        y_min = y_min_no_boundary - y_size_boundary_top
                        if y_min < 0:
                            y_min = 0
                            y_size_boundary_top = y_min_no_boundary
                        # maximum y without boundary
                        y_max_no_boundary = (
                                y_min_no_boundary + y_size_no_boundary)
                        if y_max_no_boundary >= raster_y_size:
                            y_max_no_boundary = raster_y_size
                            y_size_boundary_bottom = 0
                            y_size_no_boundary = (
                                    y_max_no_boundary - y_min_no_boundary)
                        # maximum y with boundary
                        y_max = y_max_no_boundary + y_size_boundary_bottom
                        if y_max >= raster_y_size:
                            y_max = raster_y_size
                            y_size_boundary_bottom = (
                                    raster_y_size - y_max_no_boundary)
                        y_size = y_max - y_min
                        sections_list.append(
                            processor.RasterSection(
                                x_min=0, y_min=y_min, x_max=block_size_x,
                                y_max=y_max, x_size=block_size_x,
                                y_size=y_size,
                                y_min_no_boundary=y_min_no_boundary,
                                y_max_no_boundary=y_max_no_boundary,
                                y_size_boundary_top=y_size_boundary_top,
                                y_size_boundary_bottom=y_size_boundary_bottom,
                                y_size_no_boundary=y_size_no_boundary
                            )
                        )
                        # range variables
                        y_size_range += y_size
                        y_size_no_boundary_range += y_size_no_boundary
                        y_min_no_boundary_range_list.append(y_min_no_boundary)
                        y_max_no_boundary_range_list.append(y_max_no_boundary)
                        # keep the first boundary
                        if y_size_boundary_top_range is None:
                            y_size_boundary_top_range = y_size_boundary_top
                        # keep the last one boundary
                        y_max_range = y_max
                        y_size_boundary_bottom_range = y_size_boundary_bottom
                # separate bands processes
                else:
                    # section y size
                    y_size = block_size_y
                    # adapt to raster y size
                    if y + y_size > raster_y_size:
                        y_size = raster_y_size - y
                    sections = processor.RasterSection(
                        x_min=0, y_min=y, x_max=block_size_x, y_max=y + y_size,
                        x_size=block_size_x, y_size=y_size
                    )
                    pieces.append(
                        processor.RasterPiece(
                            section_list=sections, x_min=0, y_min=y,
                            x_max=block_size_x, y_max=y + y_size,
                            x_size=block_size_x, y_size=y_size
                        )
                    )
                    y_min_no_boundary_range_list.append(y)
                    y_max_no_boundary_range_list.append(y + y_size)
            # process with sections
            if not separate_bands:
                pieces.append(
                    processor.RasterPiece(
                        section_list=sections_list, x_min=0, y_min=y_min_range,
                        x_max=block_size_x, y_max=y_max_range,
                        x_size=block_size_x, y_size=y_size_range,
                        y_min_no_boundary=min(y_min_no_boundary_range_list),
                        y_max_no_boundary=max(y_max_no_boundary_range_list),
                        y_size_boundary_top=y_size_boundary_top_range,
                        y_size_no_boundary=y_size_no_boundary_range,
                        y_size_boundary_bottom=y_size_boundary_bottom_range
                    )
                )
    cfg.logger.log.debug('len(pieces): %s' % str(len(pieces)))
    return pieces


# join multiprocess table results
def _join_results(process_result):
    cfg.logger.log.debug('start')
    output = None
    # get parallel dictionary result
    for p in sorted(process_result):
        if output is None:
            output = process_result[p]
        else:
            output = rfn.stack_arrays(
                (output, process_result[p]), asrecarray=True, usemask=False
            )
    gc.collect()
    cfg.logger.log.debug('end')
    return output



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.

"""
Tools to manage raster writing and reading
"""

import os

import numpy as np

from remotior_sensus.core import configurations as cfg
from remotior_sensus.util import files_directories, read_write_files

try:
    if cfg.gdal_path is not None:
        os.add_dll_directory(cfg.gdal_path)
except Exception as error:
    cfg.logger.log.error(str(error))
try:
    from osgeo import ogr
    from osgeo import osr
except Exception as error:
    cfg.logger.log.error(str(error))
try:
    from osgeo import gdal
except Exception as error:
    cfg.logger.log.error(str(error))


# get GDAL version
def get_gdal_version():
    v = gdal.VersionInfo('RELEASE_NAME').split('.')
    cfg.logger.log.debug('gdal version: %s' % v)
    return v


# check if file is raster or vector and return crs
def raster_or_vector_input(path):
    cfg.logger.log.debug('start')
    vector = False
    raster = False
    crs = None
    # try vector
    l_p = ogr.Open(path)
    # if raster
    if l_p is None:
        raster = True
        l_p = gdal.Open(path, gdal.GA_ReadOnly)
        if l_p is None:
            raster = False
        else:
            try:
                # check projections
                crs = l_p.GetProjection()
                crs = crs.replace(' ', '')
                if len(crs) == 0:
                    crs = None
            except Exception as err:
                crs = None
                raster = False
                cfg.logger.log.error(str(err))
    # if vector
    else:
        vector = True
        layer = l_p.GetLayer()
        # check projection
        proj = layer.GetSpatialRef()
        try:
            crs = proj.ExportToWkt()
            crs = crs.replace(' ', '')
            if len(crs) == 0:
                crs = None
        except Exception as err:
            vector = False
            crs = None
            cfg.logger.log.error(str(err))
    return vector, raster, crs


# number of raster bands
def get_number_bands(path):
    cfg.logger.log.debug('start')
    try:
        r_d = gdal.Open(path, gdal.GA_ReadOnly)
        number = r_d.RasterCount
    except Exception as err:
        number = None
        cfg.logger.log.error(str(err))
    return number


# get CRS of a raster or vector
def get_crs(path):
    # try vector
    l_p = ogr.Open(path)
    # if raster
    if l_p is None:
        l_p = gdal.Open(path, gdal.GA_ReadOnly)
        if l_p is None:
            crs = None
        else:
            try:
                # check projections
                crs = l_p.GetProjection()
                crs = crs.replace(' ', '')
                if len(crs) == 0:
                    crs = None
            except Exception as err:
                crs = None
                cfg.logger.log.error(str(err))
    # if vector
    else:
        layer = l_p.GetLayer()
        # get projection
        crs = get_layer_crs(layer)
    cfg.logger.log.debug('path: %s; crs: %s' % (path, crs))
    return crs


# get CRS of a vector layer
def get_layer_crs(layer):
    # check projection
    proj = layer.GetSpatialRef()
    try:
        crs = proj.ExportToWkt()
        crs = crs.replace(' ', '')
        if len(crs) == 0:
            crs = None
    except Exception as err:
        crs = None
        cfg.logger.log.error(str(err))
    return crs


# get spatial reference from wkt
def get_spatial_reference(input_projection):
    spatial_reference = osr.SpatialReference(wkt=input_projection)
    return spatial_reference


# compare two crs
def compare_crs(first_crs, second_crs):
    cfg.logger.log.debug(
        'first_crs: %s; second_crs: %s' % (first_crs, second_crs)
    )
    try:
        first_sr = osr.SpatialReference()
        first_sr.ImportFromWkt(first_crs)
        second_sr = osr.SpatialReference()
        second_sr.ImportFromWkt(second_crs)
        if first_sr.IsSame(second_sr) == 1:
            same = True
        else:
            same = False
        cfg.logger.log.debug('same: %s' % same)
        return same
    except Exception as err:
        cfg.logger.log.error(str(err))
        return False


# raster information
def raster_info(path):
    r_d = gdal.Open(path, gdal.GA_ReadOnly)
    if r_d is None:
        cfg.logger.log.error('raster: %s' % path)
        return False
    # x pixel count
    x_count = r_d.RasterXSize
    # y pixel count
    y_count = r_d.RasterYSize
    # geo transformation
    gt = r_d.GetGeoTransform()
    band = r_d.GetRasterBand(1)
    # nodata value
    nd = band.GetNoDataValue()
    # offset and scale
    offset = band.GetOffset()
    scale = band.GetScale()
    data_type = gdal.GetDataTypeName(band.DataType)
    unit = None
    # crs
    try:
        crs = r_d.GetProjection()
        if len(crs) == 0:
            crs = None
        else:
            crs_sr = osr.SpatialReference(wkt=crs)
            if crs_sr.IsProjected:
                unit = crs_sr.GetAttrValue('unit')
    except Exception as err:
        crs = None
        cfg.logger.log.error(str(err))
    # band number and block size
    number_of_bands = r_d.RasterCount
    block_size = band.GetBlockSize()
    info = [gt, crs, unit, [x_count, y_count], nd, number_of_bands, block_size,
            [scale, offset], data_type]
    cfg.logger.log.debug(
        '{} :{}'.format(
            path, [gt, unit, [x_count, y_count], nd, number_of_bands,
                   block_size, [scale, offset], data_type, crs]
        )
    )
    return info


# raster no data value
def raster_nodata_value(path):
    r_d = gdal.Open(path, gdal.GA_ReadOnly)
    band = r_d.GetRasterBand(1)
    nd = band.GetNoDataValue()
    return nd


# get image geotransformation
def image_geotransformation(path):
    cfg.logger.log.debug('path: %s' % path)
    # raster extent and pixel size
    (gt, r_p, unit, xy_count, nd, number_of_bands, block_size, scale_offset,
     data_type) = raster_info(path)
    left = gt[0]
    top = gt[3]
    p_x = gt[1]
    p_y = abs(gt[5])
    right = gt[0] + gt[1] * xy_count[0]
    bottom = gt[3] + gt[5] * xy_count[1]
    c_rsr = osr.SpatialReference(wkt=r_p)
    if c_rsr.IsProjected:
        un = c_rsr.GetAttrValue('unit')
    else:
        un = None
    cfg.logger.log.debug(
        'left: %s; right: %s; top: %s; bottom: %s; p_x: %s; p_y: %s; r_p: '
        '%s; un: %s'
        % (left, top, right, bottom, p_x, p_y, r_p.replace(' ', ''), un)
    )
    return left, top, right, bottom, p_x, p_y, r_p.replace(' ', ''), un


# create raster from a reference raster
def create_raster_from_reference(
        path, band_number, output_raster_list, nodata_value=None,
        driver='GTiff', gdal_format='Float32', compress=False,
        compress_format='DEFLATE21', projection=None, geo_transform=None,
        constant_value=None, x_size=None,
        y_size=None, scale=None, offset=None
):
    cfg.logger.log.debug('path: %s' % path)
    # open input_raster with GDAL
    gdal_raster_ref = gdal.Open(path, gdal.GA_ReadOnly)
    if gdal_format == 'Float64':
        gdal_format = gdal.GDT_Float64
    elif gdal_format == 'Float32':
        gdal_format = gdal.GDT_Float32
    elif gdal_format == 'Int32':
        gdal_format = gdal.GDT_Int32
    elif gdal_format == 'UInt32':
        gdal_format = gdal.GDT_UInt32
    elif gdal_format == 'Int16':
        gdal_format = gdal.GDT_Int16
    elif gdal_format == 'UInt16':
        gdal_format = gdal.GDT_UInt16
    elif gdal_format == 'Byte':
        gdal_format = gdal.GDT_Byte
    for o in output_raster_list:
        if driver == 'GTiff':
            if not o.lower().endswith(cfg.tif_suffix):
                o += cfg.tif_suffix
        # pixel size and origin from reference
        if projection is None:
            r_p = gdal_raster_ref.GetProjection()
        else:
            r_p = projection
        if geo_transform is None:
            r_gt = gdal_raster_ref.GetGeoTransform()
        else:
            r_gt = geo_transform
        t_d = gdal.GetDriverByName(driver)
        if x_size is None:
            c = gdal_raster_ref.RasterXSize
        else:
            c = x_size
        if y_size is None:
            r = gdal_raster_ref.RasterYSize
        else:
            r = y_size
        if not compress:
            out_raster = t_d.Create(o, c, r, band_number, gdal_format)
        elif compress == 'DEFLATE21':
            out_raster = t_d.Create(
                o, c, r, band_number, gdal_format,
                options=['COMPRESS=DEFLATE', 'PREDICTOR=2',
                         'ZLEVEL=1']
            )
        else:
            out_raster = t_d.Create(
                o, c, r, band_number, gdal_format,
                ['COMPRESS=%s' % compress_format]
            )
        if out_raster is None:
            cfg.logger.log.error('out_raster None')
            return False
        # set raster projection from reference
        out_raster.SetGeoTransform(r_gt)
        out_raster.SetProjection(r_p)
        if nodata_value is not None:
            for x in range(1, band_number + 1):
                _band = out_raster.GetRasterBand(x)
                _band.SetNoDataValue(nodata_value)
                _band.Fill(nodata_value)
                _band = None
        if constant_value is not None:
            for x in range(1, band_number + 1):
                _band = out_raster.GetRasterBand(x)
                _band.Fill(constant_value)
                _band = None
        if scale is not None:
            for x in range(1, band_number + 1):
                _band = out_raster.GetRasterBand(x)
                _band.SetScale(scale)
                _band = None
        if offset is not None:
            for x in range(1, band_number + 1):
                _band = out_raster.GetRasterBand(x)
                _band.SetOffset(offset)
                _band = None
    cfg.logger.log.debug('output: %s' % str(output_raster_list))
    return output_raster_list


# read a block of band as array
def read_array_block(
        gdal_band, pixel_start_column, pixel_start_row, block_columns,
        block_row, calc_data_type=None
):
    if calc_data_type is None:
        calc_data_type = np.float32
    try:
        offset = gdal_band.GetOffset()
        scale = gdal_band.GetScale()
        if offset is None:
            offset = 0.0
        if scale is None:
            scale = 1.0
    except Exception as err:
        str(err)
        offset = 0.0
        scale = 1.0
    offset = np.asarray(offset).astype(calc_data_type)
    scale = np.asarray(scale).astype(calc_data_type)
    cfg.logger.log.debug(
        'pixel_start_column: %s; pixel_start_row: %s; block_columns: %s; '
        'block_row: %s; scale: %s; offset: %s'
        % (
            pixel_start_column, pixel_start_row, block_columns, block_row,
            scale,
            offset)
    )
    try:
        a = np.asarray(
            gdal_band.ReadAsArray(
                pixel_start_column, pixel_start_row, block_columns, block_row
            ) * scale + offset
        ).astype(calc_data_type)
    except Exception as err:
        cfg.logger.log.error(str(err))
        return None
    return a


# read raster
def read_raster(raster_path, band=1, calc_data_type=None):
    (gt, r_p, unit, xy_count, nd, number_of_bands, block_size, scale_offset,
     data_type) = raster_info(raster_path)
    _r_d = gdal.Open(raster_path, gdal.GA_ReadOnly)
    r_b = _r_d.GetRasterBand(band)
    a = read_array_block(
        gdal_band=r_b, pixel_start_column=0, pixel_start_row=0,
        block_columns=xy_count[0], block_row=xy_count[1],
        calc_data_type=calc_data_type
    )
    return a


# open raster
def open_raster(raster_path, access=None):
    if access is None:
        access = gdal.GA_Update
    raster = gdal.Open(raster_path, access)
    return raster


# write raster
def write_raster(
        raster_path, x, y, data_array, nodata_value=None, scale=None,
        offset=None
):
    o_r = gdal.Open(raster_path, gdal.GA_Update)
    x_size = o_r.RasterXSize
    y_size = o_r.RasterYSize
    b_o = o_r.GetRasterBand(1)
    try:
        # it seems a GDAL issue that if scale is float the datatype is
        # converted to Float32
        if scale is not None or offset is not None:
            data_array = np.subtract(
                np.divide(data_array, scale), offset / scale
            )
            b_o.SetScale(scale)
            b_o.SetOffset(offset)
    except Exception as err:
        cfg.logger.log.error(str(err))
    if nodata_value is not None:
        b_o.SetNoDataValue(int(nodata_value))
    try:
        data_array = data_array[::, ::, 0]
    except Exception as err:
        str(err)
    cfg.logger.log.debug(
        'x_size: %s; y_size: %s; data_array.shape: %s; x: %s; y: %s'
        % (str(x_size), str(y_size), str(data_array.shape), str(x), str(y))
    )
    b_o.WriteArray(data_array, x, y)
    b_o.FlushCache()
    return raster_path


# create temporary virtual raster
def create_temporary_virtual_raster(
        input_raster_list, band_number_list=None, nodata_value=None,
        relative_to_vrt=0, intersection=True, box_coordinate_list=None,
        data_type=None, pixel_size=None, override_box_coordinate_list=None,
        grid_reference=None
):
    virtual_raster = cfg.temp.temporary_file_path(name_suffix=cfg.vrt_suffix)
    cfg.logger.log.debug(
        'virtual_raster: %s; data_type: %s'
        % (virtual_raster, data_type)
    )
    create_virtual_raster(
        input_raster_list=input_raster_list, output=virtual_raster,
        band_number_list=band_number_list, nodata_value=nodata_value,
        relative_to_vrt=relative_to_vrt, intersection=intersection,
        box_coordinate_list=box_coordinate_list, data_type=data_type,
        pixel_size=pixel_size,
        override_box_coordinate_list=override_box_coordinate_list,
        grid_reference=grid_reference
    )
    return virtual_raster


# multiband virtual raster creation
def create_virtual_raster(
        output, input_raster_list=None, band_number_list=None,
        nodata_value=None, intersection=False, relative_to_vrt=0,
        pixel_size=None, data_type=None, scale_offset_list=None,
        box_coordinate_list=None, override_box_coordinate_list=False,
        grid_reference=None, relative_extent_list=None, bandset=None,
):
    """
    :param output: output path string
    :param input_raster_list: list of raster paths
    :param band_number_list: list of lists of band numbers for each input
        raster, if None all available bands are added
    :param nodata_value: nodata value
    :param intersection: if True get minimum extent from input intersection,
        if False get maximum extent from union
    :param relative_to_vrt: integer 0 or 1, relative path of input rasters
        to vrt if 1, absolute path if 0
    :param pixel_size: optional list of pixel size x and size y
        for virtual raster output
    :param data_type: virtual raster data type
    :param scale_offset_list: optional list of scale and offset for each raster
    :param box_coordinate_list: list of boundary coordinates left top right
        bottom
    :param override_box_coordinate_list: optional with box_coordinate_list,
        if True use these absolute coordinates
    :param grid_reference: optional path of a raster used as reference grid
    :param relative_extent_list: list of xOff, yOff, xSize, ySize for source
        and destination
    :param bandset: optional BandSet

    """

    cfg.logger.log.debug('start')
    lefts = []
    rights = []
    tops = []
    bottoms = []
    p_x_sizes = []
    p_y_sizes = []
    all_band_number_list = []
    r_p = None
    r_epsg = None
    x_counts = None
    y_counts = None
    data_types = None
    x_block_sizes = None
    y_block_sizes = None
    offsets = None
    scales = None
    multiplicative_factors = None
    additive_factors = None
    nodata_values = None
    if bandset is None:
        # iterate input_raster
        for i in input_raster_list:
            # raster extent and pixel size
            (gt, r_p, unit, xy_count, nd, number_of_bands, block_size,
             scale_offset, dt) = raster_info(str(i))
            # check projections
            try:
                if r_p is not None:
                    # check projections
                    if r_epsg is None:
                        r_epsg = osr.SpatialReference()
                        r_epsg.ImportFromWkt(r_p)
                    else:
                        v_epsg = osr.SpatialReference()
                        v_epsg.ImportFromWkt(r_p)
                        if v_epsg.IsSame(r_epsg) != 1:
                            cfg.logger.log.error(
                                'v_epsg.IsSame(r_epsg): %s' % v_epsg.IsSame(
                                    r_epsg
                                )
                            )
                            return False
                else:
                    cfg.logger.log.error('error projection is None')
                    return False
            except Exception as err:
                cfg.logger.log.error(err)
                return False
            # set band number list
            if band_number_list is None:
                all_band_number_list.append(
                    list(range(1, number_of_bands + 1))
                )
            left = gt[0]
            top = gt[3]
            p_x_size = gt[1]
            p_y_size = abs(gt[5])
            right = gt[0] + gt[1] * xy_count[0]
            bottom = gt[3] + gt[5] * xy_count[1]
            # lists
            lefts.append(left)
            rights.append(right)
            tops.append(top)
            bottoms.append(bottom)
            p_x_sizes.append(p_x_size)
            p_y_sizes.append(p_y_size)
    else:
        input_raster_list = bandset.get_absolute_paths()
        lefts = bandset.get_band_attributes('left')
        tops = bandset.get_band_attributes('top')
        rights = bandset.get_band_attributes('right')
        bottoms = bandset.get_band_attributes('bottom')
        crs_s = bandset.get_band_attributes('crs')
        # check projections
        for r_p in crs_s:
            try:
                if r_p is not None:
                    # check projections
                    if r_epsg is None:
                        r_epsg = osr.SpatialReference()
                        r_epsg.ImportFromWkt(r_p)
                    else:
                        v_epsg = osr.SpatialReference()
                        v_epsg.ImportFromWkt(r_p)
                        if v_epsg.IsSame(r_epsg) != 1:
                            cfg.logger.log.error(
                                'v_epsg.IsSame(r_epsg): %s' % str(
                                    v_epsg.IsSame(r_epsg)
                                )
                            )
                            cfg.logger.log.error('error epsg')
                            return False
                else:
                    cfg.logger.log.error('error projection is None')
                    return False
            except Exception as err:
                cfg.logger.log.error(err)
                return False
        p_x_sizes = bandset.get_band_attributes('x_size')
        p_y_sizes = bandset.get_band_attributes('y_size')
        x_counts = bandset.get_band_attributes('x_count')
        y_counts = bandset.get_band_attributes('y_count')
        data_types = bandset.get_band_attributes('data_type')
        nodata_values = bandset.get_band_attributes('nodata')
        x_block_sizes = bandset.get_band_attributes('x_block_size')
        y_block_sizes = bandset.get_band_attributes('y_block_size')
        scales = bandset.get_band_attributes('scale')
        offsets = bandset.get_band_attributes('offset')
        multiplicative_factors = bandset.get_band_attributes(
            'multiplicative_factor'
        )
        additive_factors = bandset.get_band_attributes('additive_factor')
        box_coordinate_list = bandset.box_coordinate_list
        band_number_list = bandset.get_raster_band_list()
    cfg.logger.log.debug(
        'lefts: %s; rights: %s; tops: %s; bottoms: %s; p_x_sizes: %s; '
        'p_y_sizes: %s' % (
            str(lefts), str(rights), str(tops), str(bottoms), str(p_x_sizes),
            str(p_y_sizes))
    )
    # calculate boundaries
    try:
        # minimum extent of intersection
        if intersection:
            i_left = max(lefts)
            i_top = min(tops)
            i_right = min(rights)
            i_bottom = max(bottoms)
        # maximum extent of union
        else:
            i_left = min(lefts)
            i_top = max(tops)
            i_right = max(rights)
            i_bottom = min(bottoms)
    except Exception as err:
        cfg.logger.log.error(str(err))
        return False
    cfg.logger.log.debug(
        '[i_left, i_top, i_right, i_bottom]: %s' % str(
            [i_left, i_top, i_right, i_bottom]
        )
    )
    if box_coordinate_list is None:
        if pixel_size is None:
            # minimum x and y pixel size
            pixel_x_size = min(p_x_sizes)
            pixel_y_size = min(p_y_sizes)
            input_reference = input_raster_list[p_x_sizes.index(pixel_x_size)]
            # raster extent and pixel size
            (gt, p, unit, xy_count, nd, number_of_bands, block_size,
             scale_offset, dt) = raster_info(input_reference)
            r_left = gt[0]
            r_top = gt[3]
            # find raster box
            diff_left = round((i_left - r_left) / pixel_x_size) * pixel_x_size
            i_left = r_left + diff_left
            diff_top = round((i_top - r_top) / pixel_y_size) * pixel_y_size
            i_top = r_top + diff_top
            diff_right = round(
                (i_right - i_left) / pixel_x_size
            ) * pixel_x_size
            i_right = i_left + diff_right
            diff_bottom = round(
                (i_bottom - i_top) / pixel_y_size
            ) * pixel_y_size
            i_bottom = i_top + diff_bottom
            cfg.logger.log.debug(
                '[pixel_x_size, pixel_y_size]: %s; [r_left, r_top]: %s; ['
                'i_left, i_top, i_right, i_bottom]: %s'
                % (str([pixel_x_size, pixel_y_size]), str([r_left, r_top]),
                   str([i_left, i_top, i_right, i_bottom]))
            )
        else:
            pixel_x_size, pixel_y_size = pixel_size
            if pixel_x_size is None:
                # minimum x and y pixel size
                pixel_x_size = min(p_x_sizes)
            if pixel_y_size is None:
                pixel_y_size = min(p_y_sizes)
    # box coordinates to be adapted
    else:
        # override coordinates with absolute coordinates
        if override_box_coordinate_list:
            i_left = box_coordinate_list[0]
            i_top = box_coordinate_list[1]
            i_right = box_coordinate_list[2]
            i_bottom = box_coordinate_list[3]
            # x and y pixel size
            if pixel_size is None:
                pixel_x_size = min(p_x_sizes)
                pixel_y_size = min(p_y_sizes)
            else:
                pixel_x_size, pixel_y_size = pixel_size
        # find minimum extent raster
        elif grid_reference is not None:
            # raster extent and pixel size
            (gt, p, unit, xy_count, nd, number_of_bands, block_size,
             scale_offset, dt) = raster_info(grid_reference)
            r_left = gt[0]
            r_top = gt[3]
            if pixel_size is None:
                pixel_x_size = gt[1]
                pixel_y_size = abs(gt[5])
            else:
                pixel_x_size, pixel_y_size = pixel_size
            # find raster box
            diff_left = round(
                (box_coordinate_list[0] - r_left) / pixel_x_size
            ) * pixel_x_size
            i_left = r_left + diff_left
            diff_top = round(
                (box_coordinate_list[1] - r_top) / pixel_y_size
            ) * pixel_y_size
            i_top = r_top + diff_top
            diff_right = round(
                (box_coordinate_list[2] - i_left) / pixel_x_size
            ) * pixel_x_size
            i_right = i_left + diff_right
            diff_bottom = round(
                (box_coordinate_list[3] - i_top) / pixel_y_size
            ) * pixel_y_size
            i_bottom = i_top + diff_bottom
            cfg.logger.log.debug(
                '[pixel_x_size, pixel_y_size]: %s; [r_left, r_top]: %s; ['
                'i_left, i_top, i_right, i_bottom]: %s'
                % (str([pixel_x_size, pixel_y_size]), str([r_left, r_top]),
                   str([i_left, i_top, i_right, i_bottom]))
            )
        # intersection extent
        else:
            # x and y pixel size
            if pixel_size is None:
                pixel_x_size = min(p_x_sizes)
                pixel_y_size = min(p_y_sizes)
            else:
                pixel_x_size, pixel_y_size = pixel_size
            if pixel_x_size in p_x_sizes:
                input_reference = input_raster_list[
                    p_x_sizes.index(pixel_x_size)]
            else:
                input_reference = input_raster_list[
                    p_x_sizes.index(min(p_x_sizes))]
            # raster extent and pixel size
            (gt, p, unit, xy_count, nd, number_of_bands, block_size,
             scale_offset, dt) = raster_info(input_reference)
            r_left = gt[0]
            r_top = gt[3]
            # find raster box
            diff_left = round(
                (box_coordinate_list[0] - r_left) / pixel_x_size
            ) * pixel_x_size
            i_left = r_left + diff_left
            if i_left < r_left:
                i_left = r_left
            diff_top = round(
                (box_coordinate_list[1] - r_top) / pixel_y_size
            ) * pixel_y_size
            i_top = r_top + diff_top
            if i_top > r_top:
                i_top = r_top
            diff_right = round(
                (box_coordinate_list[2] - i_left) / pixel_x_size
            ) * pixel_x_size
            i_right = i_left + diff_right
            r_right = gt[0] + gt[1] * xy_count[0]
            if i_right > r_right:
                i_right = r_right
            diff_bottom = round(
                (box_coordinate_list[3] - i_top) / pixel_y_size
            ) * pixel_y_size
            i_bottom = i_top + diff_bottom
            r_bottom = gt[3] + gt[5] * xy_count[1]
            if i_bottom < r_bottom:
                i_bottom = r_bottom
    # create virtual raster
    drv = gdal.GetDriverByName('vrt')
    # number of pixels for x and y pixel size
    r_x = abs(int(round((i_right - i_left) / pixel_x_size)))
    r_y = abs(int(round((i_top - i_bottom) / pixel_y_size)))
    # create virtual raster
    files_directories.create_parent_directory(output)
    v_rast = drv.Create(output, r_x, r_y, 0)
    v_rast.SetGeoTransform((i_left, pixel_x_size, 0, i_top, 0, -pixel_y_size))
    v_rast.SetProjection(r_p)
    # set band number list
    if band_number_list is None:
        band_number_list = all_band_number_list
    cfg.logger.log.debug(
        '[r_x, r_y]: %s; band_number_list: %s' % (
            str([r_x, r_y]), str(band_number_list))
    )
    n = 0
    # iterate bands
    for raster in range(len(input_raster_list)):
        for band_number in band_number_list[raster]:
            n += 1
            cfg.logger.log.debug(
                'input_raster: %s' % str(input_raster_list[raster])
            )
            if bandset is None:
                # open input_raster
                input_raster = gdal.Open(
                    input_raster_list[raster], gdal.GA_ReadOnly
                )
                ir_x_size = input_raster.RasterXSize
                ir_y_size = input_raster.RasterYSize
                input_band = input_raster.GetRasterBand(band_number)
                # data type
                try:
                    data_type_s = gdal.GetDataTypeName(input_band.DataType)
                except Exception as err:
                    str(err)
                    data_type_s = cfg.float32_dt
                if data_type is not None:
                    try:
                        gdal_format = eval('gdal.GDT_%s' % data_type)
                    except Exception as err:
                        str(err)
                        gdal_format = input_band.DataType
                else:
                    gdal_format = input_band.DataType
                # add virtual band
                v_rast.AddBand(gdal_format)
                band = v_rast.GetRasterBand(n)
                # block size
                bsize = input_band.GetBlockSize()
                x_block = bsize[0]
                y_block = bsize[1]
                if scale_offset_list is None:
                    offs = input_band.GetOffset()
                    scl = input_band.GetScale()
                else:
                    scl, offs = scale_offset_list[raster]
                # get nodata value
                no_data = input_band.GetNoDataValue()
                # get input_raster geotransformation
                gt = input_raster.GetGeoTransform()
                p_x_size = abs(gt[1])
                p_y_size = abs(gt[5])
                left = gt[0]
                top = gt[3]
                right = gt[0] + gt[1] * ir_x_size
                bottom = gt[3] + gt[5] * ir_y_size
                # set source parameters
                if i_left <= left:
                    s_off_x = 0
                else:
                    s_off_x = round((i_left - left) / p_x_size)
                if i_top >= top:
                    s_off_y = 0
                else:
                    s_off_y = round((top - i_top) / p_y_size)
                if i_right >= right:
                    s_r_x = ir_x_size - s_off_x
                else:
                    s_r_x = ir_x_size - s_off_x - abs(
                        round((right - i_right) / p_x_size)
                    )
                if i_bottom <= bottom:
                    s_r_y = ir_y_size - s_off_y
                else:
                    s_r_y = ir_y_size - s_off_y - abs(
                        round((i_bottom - bottom) / p_y_size)
                    )
                # set destination parameters
                if i_left < left:
                    d_off_x = abs(round((left - i_left) / pixel_x_size))
                else:
                    d_off_x = 0
                if i_top > top:
                    d_off_y = abs(round((i_top - top) / pixel_y_size))
                else:
                    d_off_y = 0
                # number of x pixels
                d_r_x = round((s_r_x * p_x_size) / pixel_x_size)
                # number of y pixels
                d_r_y = round((s_r_y * p_y_size) / pixel_y_size)
                if relative_extent_list is not None:
                    s_off_x, s_off_y, s_r_x, s_r_y = relative_extent_list
                    d_off_x, d_off_y, d_r_x, d_r_y = relative_extent_list
            # BandSet
            else:
                ir_x_size = x_counts[raster]
                ir_y_size = y_counts[raster]
                # data type
                try:
                    data_type_s = data_types[raster]
                except Exception as err:
                    str(err)
                    data_type_s = cfg.float32_dt
                if data_type is not None:
                    try:
                        gdal_format = eval('gdal.GDT_%s' % data_type)
                    except Exception as err:
                        str(err)
                        gdal_format = eval('gdal.GDT_%s' % data_types[raster])
                else:
                    gdal_format = eval('gdal.GDT_%s' % data_types[raster])
                # add virtual band
                v_rast.AddBand(gdal_format)
                band = v_rast.GetRasterBand(n)
                # block size
                x_block = x_block_sizes[raster]
                y_block = y_block_sizes[raster]
                if scale_offset_list is None:
                    if offsets[raster] == cfg.nodata_val_Int64:
                        offs = None
                    else:
                        offs = offsets[raster]
                    if scales[raster] == cfg.nodata_val_Int64:
                        scl = None
                    else:
                        scl = scales[raster]
                else:
                    scl, offs = scale_offset_list[raster]
                if scl is not None:
                    scl = multiplicative_factors[raster] * scl
                if offs is not None:
                    offs = additive_factors[raster] + offs
                # get nodata value
                no_data = nodata_values[raster]
                p_x_size = p_x_sizes[raster]
                p_y_size = p_y_sizes[raster]
                left = lefts[raster]
                top = tops[raster]
                right = rights[raster]
                bottom = bottoms[raster]
                # set source parameters
                if i_left <= left:
                    s_off_x = 0
                else:
                    s_off_x = round((i_left - left) / p_x_size)
                if i_top >= top:
                    s_off_y = 0
                else:
                    s_off_y = round((top - i_top) / p_y_size)
                if i_right >= right:
                    s_r_x = ir_x_size - s_off_x
                else:
                    s_r_x = ir_x_size - s_off_x - abs(
                        round((right - i_right) / p_x_size)
                    )
                if i_bottom <= bottom:
                    s_r_y = ir_y_size - s_off_y
                else:
                    s_r_y = ir_y_size - s_off_y - abs(
                        round((i_bottom - bottom) / p_y_size)
                    )
                # set destination parameters
                if i_left < left:
                    d_off_x = abs(round((left - i_left) / pixel_x_size))
                else:
                    d_off_x = 0
                if i_top > top:
                    d_off_y = abs(round((i_top - top) / pixel_y_size))
                else:
                    d_off_y = 0
                # number of x pixels
                d_r_x = round((s_r_x * p_x_size) / pixel_x_size)
                # number of y pixels
                d_r_y = round((s_r_y * p_y_size) / pixel_y_size)
                if relative_extent_list is not None:
                    s_off_x, s_off_y, s_r_x, s_r_y = relative_extent_list
                    d_off_x, d_off_y, d_r_x, d_r_y = relative_extent_list
            cfg.logger.log.debug(
                '[left, top, right, bottom]: %s; [s_off_x, s_off_y, s_r_x, '
                's_r_y]: %s;[d_off_x, d_off_y, d_r_x, '
                'd_r_y]: %s; data_type_s: %s; nodata_value: %s; no_data: %s'
                % (str([left, top, right, bottom]),
                   str([s_off_x, s_off_y, s_r_x, s_r_y]),
                   str([d_off_x, d_off_y, d_r_x, d_r_y]), str(data_type_s),
                   str(nodata_value), str(no_data))
            )
            try:
                # check path
                if relative_to_vrt == 1:
                    source_path = files_directories.file_name(
                        input_raster_list[raster], True
                    )
                else:
                    source_path = input_raster_list[raster].replace(
                        '//', '/'
                    ).replace('https:/', 'https://').replace(
                        'http:/', 'http://'
                    )
                # set metadata xml
                if no_data is None:
                    xml = '''
                <ComplexSource>
                    <SourceFilename relative_to_vrt="%i">%s</SourceFilename>
                    <SourceBand>%i</SourceBand>
                    <SourceProperties RasterXSize="%i" RasterYSize="%i" 
                    DataType="%s" BlockXSize="%i" BlockYSize="%i" />
                    <SrcRect xOff="%i" yOff="%i" xSize="%i" ySize="%i" />
                    <DstRect xOff="%i" yOff="%i" xSize="%i" ySize="%i" />
                </ComplexSource>
                    '''
                    source = xml % (relative_to_vrt, source_path,
                                    band_number,
                                    ir_x_size, ir_y_size, data_type_s, x_block,
                                    y_block,
                                    s_off_x, s_off_y, s_r_x, s_r_y,
                                    d_off_x, d_off_y, d_r_x, d_r_y)
                else:
                    xml = '''
                <ComplexSource>
                    <SourceFilename relative_to_vrt="%i">%s</SourceFilename>
                    <SourceBand>%i</SourceBand>
                    <SourceProperties RasterXSize="%i" RasterYSize="%i" 
                    DataType="%s" BlockXSize="%i" BlockYSize="%i" />
                    <SrcRect xOff="%i" yOff="%i" xSize="%i" ySize="%i" />
                    <DstRect xOff="%i" yOff="%i" xSize="%i" ySize="%i" />
                    <NODATA>%i</NODATA>
                </ComplexSource>
                    '''
                    source = xml % (relative_to_vrt, source_path, band_number,
                                    ir_x_size, ir_y_size, data_type_s, x_block,
                                    y_block, s_off_x, s_off_y, s_r_x, s_r_y,
                                    d_off_x, d_off_y, d_r_x, d_r_y, no_data)
                band.SetMetadataItem(
                    'ComplexSource', source, 'new_vrt_sources'
                )
                if nodata_value is True:
                    if no_data is not None:
                        band.SetNoDataValue(int(no_data))
                elif not nodata_value:
                    try:
                        band.SetNoDataValue(int(no_data))
                    except Exception as err:
                        str(err)
                else:
                    try:
                        band.SetNoDataValue(int(nodata_value))
                    except Exception as err:
                        str(err)
                if offs is not None:
                    band.SetOffset(offs)
                if scl is not None:
                    band.SetScale(scl)
            except Exception as err:
                cfg.logger.log.error(str(err))
    cfg.logger.log.debug('end; output: %s' % str(output))
    return str(output)


# simplified virtual raster creation for mosaic
def create_virtual_raster_2_mosaic(
        input_raster_list, output, band_number_list=None, src_nodata=None,
        dst_nodata=False, relative_to_vrt=0, data_type=None,
        box_coordinate_list=None, override_box_coordinate_list=False,
        pixel_size=None, grid_reference=None, scale_offset_list=None
):
    cfg.logger.log.debug('start')
    lefts = []
    rights = []
    tops = []
    bottoms = []
    p_x_sizes = []
    p_y_sizes = []
    r_p = None
    r_epsg = None
    dt0 = None
    # iterate input_raster
    for i in input_raster_list:
        # raster extent and pixel size
        (gt, r_p, unit, xy_count, nd, number_of_bands, block_size,
         scale_offset, dt) = raster_info(i)
        # data type
        if dt0 is None:
            dt0 = dt
        # check projections
        try:
            if r_p is not None:
                # check projections
                if r_epsg is None:
                    r_epsg = osr.SpatialReference()
                    r_epsg.ImportFromWkt(r_p)
                else:
                    v_epsg = osr.SpatialReference()
                    v_epsg.ImportFromWkt(r_p)
                    if v_epsg.IsSame(r_epsg) != 1:
                        cfg.logger.log.error('error epsg')
                        return False
            else:
                cfg.logger.log.error('error projection')
                return False
        except Exception as err:
            cfg.logger.log.error(err)
            return False
        left = gt[0]
        top = gt[3]
        p_x_size = gt[1]
        p_y_size = abs(gt[5])
        right = gt[0] + gt[1] * xy_count[0]
        bottom = gt[3] + gt[5] * xy_count[1]
        # lists
        lefts.append(left)
        rights.append(right)
        tops.append(top)
        bottoms.append(bottom)
        p_x_sizes.append(p_x_size)
        p_y_sizes.append(p_y_size)
    # calculate boundaries
    try:
        # maximum extent of union
        i_left = min(lefts)
        i_top = max(tops)
        i_right = max(rights)
        i_bottom = min(bottoms)
    except Exception as err:
        cfg.logger.log.error(str(err))
        return False
    pixel_x_size, pixel_y_size = None, None
    if box_coordinate_list is None:
        if pixel_size is None:
            # minimum x and y pixel size
            pixel_x_size = min(p_x_sizes)
            pixel_y_size = min(p_y_sizes)
            input_reference = input_raster_list[p_x_sizes.index(pixel_x_size)]
            # raster extent and pixel size
            (gt, p, unit, xy_count, nd, number_of_bands, block_size,
             scale_offset, dt) = raster_info(input_reference)
            r_left = gt[0]
            r_top = gt[3]
            # find raster box
            diff_left = round((i_left - r_left) / pixel_x_size) * pixel_x_size
            i_left = r_left + diff_left
            diff_top = round((i_top - r_top) / pixel_y_size) * pixel_y_size
            i_top = r_top + diff_top
            diff_right = round(
                (i_right - i_left) / pixel_x_size
            ) * pixel_x_size
            i_right = i_left + diff_right
            diff_bottom = round(
                (i_bottom - i_top) / pixel_y_size
            ) * pixel_y_size
            i_bottom = i_top + diff_bottom
        else:
            pixel_x_size, pixel_y_size = pixel_size
    # box coordinates to be adapted
    else:
        # override coordinates with absolute coordinates
        if override_box_coordinate_list:
            i_left = box_coordinate_list[0]
            i_top = box_coordinate_list[1]
            i_right = box_coordinate_list[2]
            i_bottom = box_coordinate_list[3]
        # find minimum extent raster
        elif grid_reference is not None:
            # raster extent and pixel size
            (gt, p, unit, xy_count, nd, number_of_bands, block_size,
             scale_offset, dt) = raster_info(grid_reference)
            r_left = gt[0]
            r_top = gt[3]
            if pixel_size is None:
                pixel_x_size = gt[1]
                pixel_y_size = abs(gt[5])
            else:
                pixel_x_size, pixel_y_size = pixel_size
            # find raster box
            diff_left = round(
                (box_coordinate_list[0] - r_left) / pixel_x_size
            ) * pixel_x_size
            i_left = r_left + diff_left
            diff_top = round(
                (box_coordinate_list[1] - r_top) / pixel_y_size
            ) * pixel_y_size
            i_top = r_top + diff_top
            diff_right = round(
                (box_coordinate_list[2] - i_left) / pixel_x_size
            ) * pixel_x_size
            i_right = i_left + diff_right
            diff_bottom = round(
                (box_coordinate_list[3] - i_top) / pixel_y_size
            ) * pixel_y_size
            i_bottom = i_top + diff_bottom
        # intersection extent
        else:
            if i_left < box_coordinate_list[0]:
                i_left = box_coordinate_list[0]
            if i_top > box_coordinate_list[1]:
                i_top = box_coordinate_list[1]
            if i_right > box_coordinate_list[2]:
                i_right = box_coordinate_list[2]
            if i_bottom < box_coordinate_list[3]:
                i_bottom = box_coordinate_list[3]
    # create virtual raster
    drv = gdal.GetDriverByName('vrt')
    # number of pixels for x and y pixel size
    r_x = abs(int(round((i_right - i_left) / pixel_x_size)))
    r_y = abs(int(round((i_top - i_bottom) / pixel_y_size)))
    # create virtual raster
    files_directories.create_parent_directory(output)
    v_rast = drv.Create(output, r_x, r_y, 0)
    v_rast.SetGeoTransform((i_left, pixel_x_size, 0, i_top, 0, -pixel_y_size))
    v_rast.SetProjection(r_p)
    if data_type is not None:
        try:
            gdal_format = eval('gdal.GDT_%s' % data_type)
        except Exception as err:
            str(err)
            gdal_format = eval('gdal.GDT_%s' % dt0)
    else:
        gdal_format = eval('gdal.GDT_%s' % dt0)
    # add virtual band
    v_rast.AddBand(gdal_format)
    band = v_rast.GetRasterBand(1)
    # iterate bands
    for b in range(len(input_raster_list)):
        # open input_raster
        if band_number_list is None:
            band_number = 1
        else:
            band_number = band_number_list[b]
        # open input_raster
        input_raster = gdal.Open(input_raster_list[b], gdal.GA_ReadOnly)
        input_band = input_raster.GetRasterBand(band_number)
        if scale_offset_list is None:
            offs = input_band.GetOffset()
            scl = input_band.GetScale()
        else:
            scl, offs = scale_offset_list[b]
        # get nodata value
        no_data = input_band.GetNoDataValue()
        i_r_x = input_raster.RasterXSize
        i_r_y = input_raster.RasterYSize
        # data type
        try:
            data_type_s = gdal.GetDataTypeName(input_band.DataType)
        except Exception as err:
            str(err)
            data_type_s = cfg.float32_dt
        if src_nodata is None:
            src_nodata = no_data
        # block size
        bsize = input_band.GetBlockSize()
        x_block = bsize[0]
        y_block = bsize[1]
        # get input_raster geotransformation
        gt = input_raster.GetGeoTransform()
        p_x_size = abs(gt[1])
        p_y_size = abs(gt[5])
        left = gt[0]
        top = gt[3]
        right = gt[0] + gt[1] * i_r_x
        bottom = gt[3] + gt[5] * i_r_y
        # set source parameters
        if i_left <= left:
            s_off_x = 0
        else:
            s_off_x = round((i_left - left) / p_x_size) * p_x_size
        if i_top >= top:
            s_off_y = 0
        else:
            s_off_y = round((i_top - top) / p_y_size) * p_y_size
        if i_right >= right:
            s_r_x = i_r_x - s_off_x
        else:
            s_r_x = i_r_x - s_off_x - abs(
                round((right - i_right) / p_x_size) * p_x_size
            )
        if i_bottom <= bottom:
            s_r_y = i_r_y - s_off_y
        else:
            s_r_y = i_r_y - s_off_y - abs(
                round((i_bottom - bottom) / p_y_size) * p_y_size
            )
        # set destination parameters
        if i_left < left:
            d_off_x = abs(round((left - i_left) / pixel_x_size))
        else:
            d_off_x = 0
        if i_top > top:
            d_off_y = abs(round((i_top - top) / pixel_y_size))
        else:
            d_off_y = 0
        # number of x pixels
        d_r_x = round((s_r_x * p_x_size) / pixel_x_size)
        # number of y pixels
        d_r_y = round((s_r_y * p_y_size) / pixel_y_size)
        try:
            # check path
            if relative_to_vrt == 1:
                source_path = files_directories.file_name(
                    input_raster_list[b], True
                )
            else:
                source_path = input_raster_list[b].replace('//', '/')
            # set metadata xml
            if src_nodata is None:
                xml = '''
                <ComplexSource>
                    <SourceFilename relative_to_vrt="%i">%s</SourceFilename>
                    <SourceBand>%i</SourceBand>
                    <SourceProperties RasterXSize="%i" RasterYSize="%i" 
                    DataType="%s" BlockXSize="%i" BlockYSize="%i" />
                    <SrcRect xOff="%i" yOff="%i" xSize="%i" ySize="%i" />
                    <DstRect xOff="%i" yOff="%i" xSize="%i" ySize="%i" />
                </ComplexSource>
                '''
                source = xml % (relative_to_vrt, source_path, band_number,
                                i_r_x, i_r_y, data_type_s, x_block, y_block,
                                s_off_x, s_off_y, s_r_x, s_r_y,
                                d_off_x, d_off_y, d_r_x, d_r_y)
                band.SetMetadataItem(
                    'ComplexSource', source, 'new_vrt_sources'
                )
            else:
                xml = '''
                <ComplexSource>
                    <SourceFilename relative_to_vrt="%i">%s</SourceFilename>
                    <SourceBand>%i</SourceBand>
                    <SourceProperties RasterXSize="%i" RasterYSize="%i" 
                    DataType="%s" BlockXSize="%i" BlockYSize="%i" />
                    <SrcRect xOff="%i" yOff="%i" xSize="%i" ySize="%i" />
                    <DstRect xOff="%i" yOff="%i" xSize="%i" ySize="%i" />
                    <NODATA>%i</NODATA>
                </ComplexSource>
                '''
                source = xml % (relative_to_vrt, source_path, band_number,
                                i_r_x, i_r_y, data_type_s, x_block, y_block,
                                s_off_x, s_off_y, s_r_x, s_r_y,
                                d_off_x, d_off_y, d_r_x, d_r_y, src_nodata)
                band.SetMetadataItem(
                    'ComplexSource', source, 'new_vrt_sources'
                )
            if dst_nodata is True:
                band.SetNoDataValue(int(src_nodata))
            elif not dst_nodata:
                pass
            else:
                try:
                    band.SetNoDataValue(int(dst_nodata))
                except Exception as err:
                    str(err)
                    band.SetNoDataValue(int(no_data))
            if offs is not None:
                band.SetOffset(offs)
            if scl is not None:
                band.SetScale(scl)
        except Exception as err:
            cfg.logger.log.error(str(err))
    cfg.logger.log.debug('end; virtual raster: %s' % str(output))
    return str(output)


# overwrite text to force relative to vrt
def force_relative_to_vrt(file_path):
    text = read_write_files.open_text_file(file_path)
    text = text.replace('relativeToVRT="0"', 'relativeToVRT="1"')
    read_write_files.write_file(text, file_path)
    return file_path


# project point coordinates
def project_point_coordinates(
        point_x, point_y, input_coordinates, output_coordinates
):
    cfg.logger.log.debug('start')
    try:
        # required by GDAL 3 coordinate order
        try:
            input_coordinates.SetAxisMappingStrategy(
                osr.OAMS_TRADITIONAL_GIS_ORDER
            )
            output_coordinates.SetAxisMappingStrategy(
                osr.OAMS_TRADITIONAL_GIS_ORDER
            )
        except Exception as err:
            str(err)
        # coordinate transformation
        c_t = osr.CoordinateTransformation(
            input_coordinates, output_coordinates
        )
        point_t = ogr.Geometry(ogr.wkbPoint)
        point_t.AddPoint(point_x, point_y)
        point_t.Transform(c_t)
        cfg.logger.log.debug(
            '[point_t.GetX(), point_t.GetY()]: %s' % str(
                [point_t.GetX(), point_t.GetY()]
            )
        )
        return [point_t.GetX(), point_t.GetY()]
    except Exception as err:
        cfg.logger.log.error(str(err))
        return False


# reproject vector
def reproject_vector(
        input_vector, output, input_epsg=None, output_epsg=None,
        vector_type='wkbMultiPolygon',
        output_drive=None
):
    cfg.logger.log.debug('start')
    # input spatial reference
    input_sr = osr.SpatialReference()
    if input_epsg is None:
        l_p = ogr.Open(input_vector)
        layer = l_p.GetLayer()
        proj = layer.GetSpatialRef()
        try:
            crs = proj.ExportToWkt()
            input_sr = crs.replace(' ', '')
            if len(input_sr) == 0:
                cfg.logger.log.error('Error input vector')
                return False
        except Exception as err:
            cfg.logger.log.error(str(err))
            return False
    else:
        # input EPSG or projection
        try:
            input_sr.ImportFromEPSG(input_epsg)
        except Exception as err:
            str(err)
            input_sr = input_epsg
    # output spatial reference
    output_sr = osr.SpatialReference()
    try:
        output_sr.ImportFromEPSG(output_epsg)
    except Exception as err:
        str(err)
        output_sr = output_epsg
    # required by GDAL 3 coordinate order
    try:
        input_sr.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
        output_sr.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    except Exception as err:
        str(err)
    # coordinate transformation
    c_t = osr.CoordinateTransformation(input_sr, output_sr)
    # open input vector
    i_vector = ogr.Open(input_vector)
    i_layer = i_vector.GetLayer()
    # create output vector
    if output_drive is None:
        if files_directories.file_extension(
                output, lower=True
        ) == cfg.shp_suffix:
            output_drive = 'ESRI Shapefile'
        else:
            output_drive = 'GPKG'
    i_driver = ogr.GetDriverByName(output_drive)
    o_source = i_driver.CreateDataSource(output)
    name = files_directories.file_name(output, suffix=False)
    if vector_type == 'wkbMultiPolygon':
        o_layer = o_source.CreateLayer(name, output_sr, ogr.wkbMultiPolygon)
    elif vector_type == 'wkbPoint':
        o_layer = o_source.CreateLayer(name, output_sr, ogr.wkbPoint)
    else:
        cfg.logger.log.error('Error vector type')
        return False
    i_layer_definition = i_layer.GetLayerDefn()
    # copy fields
    for i in range(i_layer_definition.GetFieldCount()):
        f_definition = i_layer_definition.GetFieldDefn(i)
        o_layer.CreateField(f_definition)
    o_layer_definition = o_layer.GetLayerDefn()
    # field_value count
    o_field_count = o_layer_definition.GetFieldCount()
    # iterate input features
    i_feature = i_layer.GetNextFeature()
    while i_feature:
        # get geometry
        geom = i_feature.GetGeometryRef()
        # project feature
        geom.Transform(c_t)
        o_feature = ogr.Feature(o_layer_definition)
        o_feature.SetGeometry(geom)
        for i in range(o_field_count):
            field_name = o_layer_definition.GetFieldDefn(i).GetNameRef()
            field_value = i_feature.GetField(i)
            o_feature.SetField(field_name, field_value)
        o_layer.CreateFeature(o_feature)
        o_feature.Destroy()
        i_feature.Destroy()
        i_feature = i_layer.GetNextFeature()
    # close files
    i_vector.Destroy()
    o_source.Destroy()
    cfg.logger.log.debug('output: %s' % output)
    return output


# get layer extent
def get_layer_extent(layer_path):
    _temp_vector = ogr.Open(layer_path)
    _temp_layer = _temp_vector.GetLayer()
    min_x, max_x, min_y, max_y = _temp_layer.GetExtent()
    _temp_layer = None
    _temp_vector = None
    return min_x, max_x, min_y, max_y


# convert reference layer to raster based on the resolution of a raster
def vector_to_raster(
        vector_path=None, output_path=None, field_name=None,
        reference_raster_path=None, all_touched=None,
        output_format='GTiff', burn_values=None, attribute_filter=None,
        extent=None, nodata_value=0,
        background_value=0, compress=False, compress_format='DEFLATE21',
        vector_layer=None
):
    cfg.logger.log.debug('start')
    if vector_path is not None:
        vector_crs = get_crs(vector_path)
    elif vector_layer is not None:
        vector_crs = get_layer_crs(vector_layer)
    else:
        cfg.logger.log.error('input vector')
        return False
    (gt, reference_crs, unit, xy_count, nd, number_of_bands, block_size,
     scale_offset, data_type) = raster_info(reference_raster_path)
    orig_x = gt[0]
    orig_y = gt[3]
    # number of x pixels
    grid_columns = xy_count[0]
    # number of y pixels
    grid_rows = xy_count[1]
    # check crs
    same_crs = compare_crs(reference_crs, vector_crs)
    if vector_path is not None:
        if not same_crs:
            input_vector = cfg.temp.temporary_file_path(
                name_suffix=files_directories.file_extension(vector_path)
            )
            reproject_vector(
                vector_path, input_vector, input_epsg=vector_crs,
                output_epsg=reference_crs
            )
        else:
            input_vector = vector_path
        # open input vector
        vector = ogr.Open(input_vector)
        # get layer
        try:
            v_layer = vector.GetLayer()
        except Exception as err:
            cfg.logger.log.error(err)
            return False
    elif vector_layer is not None:
        if not same_crs:
            cfg.logger.log.error('different crs')
            return False
        else:
            v_layer = vector_layer
    else:
        v_layer = None
    # attribute filter
    if attribute_filter is not None:
        v_layer.SetAttributeFilter(attribute_filter)
        d = ogr.GetDriverByName('MEMORY')
        d_s = d.CreateDataSource('memData')
        layer_copy = d_s.CopyLayer(v_layer, d_s.GetName(), ['OVERWRITE=YES'])
        min_x, max_x, min_y, max_y = layer_copy.GetExtent()
        v_layer = layer_copy
    else:
        min_x, max_x, min_y, max_y = v_layer.GetExtent()
    # calculate minimum extent
    if not extent:
        orig_x = gt[0] + gt[1] * int(round((min_x - gt[0]) / gt[1]))
        orig_y = gt[3] + gt[5] * int(round((max_y - gt[3]) / gt[5]))
        grid_columns = abs(int(round((max_x - min_x) / gt[1])))
        grid_rows = abs(int(round((max_y - min_y) / gt[5])))
    driver = gdal.GetDriverByName(output_format)
    temporary_grid = cfg.temp.temporary_raster_path(extension=cfg.tif_suffix)
    # create raster _grid
    _grid = driver.Create(
        temporary_grid, grid_columns, grid_rows, 1, gdal.GDT_Float32
    )
    if _grid is None:
        _grid = driver.Create(
            temporary_grid, grid_columns, grid_rows, 1, gdal.GDT_Int16
        )
    if _grid is None:
        cfg.logger.log.error('error output raster')
        return False
    try:
        _grid.GetRasterBand(1)
    except Exception as err:
        cfg.logger.log.error(err)
        return False
    # set raster projection from reference
    _grid.SetGeoTransform([orig_x, gt[1], 0, orig_y, 0, gt[5]])
    _grid.SetProjection(reference_crs)
    _grid = None
    # create output raster
    create_raster_from_reference(
        path=temporary_grid, band_number=1, output_raster_list=[output_path],
        nodata_value=nodata_value, driver='GTiff',
        gdal_format=cfg.raster_data_type, compress=compress,
        compress_format=compress_format, constant_value=background_value
    )
    # convert reference layer to raster
    output_raster = gdal.Open(output_path, gdal.GA_Update)
    if all_touched is None:
        if burn_values is None:
            o_c = gdal.RasterizeLayer(
                output_raster, [1], v_layer, options=[
                    'ATTRIBUTE=%s' % str(field_name)]
            )
        else:
            o_c = gdal.RasterizeLayer(
                output_raster, [1], v_layer, burn_values=[burn_values]
            )
    else:
        if burn_values is None:
            o_c = gdal.RasterizeLayer(
                output_raster, [1], v_layer,
                options=['ATTRIBUTE=%s' % str(field_name),
                         'all_touched=TRUE']
            )
        else:
            o_c = gdal.RasterizeLayer(
                output_raster, [1], v_layer, burn_values=[burn_values],
                options=['all_touched=TRUE']
            )
    cfg.logger.log.debug('gdal rasterize: %s' % str(o_c))
    return output_path


# merge all layers to new layer
def merge_all_layers(
        input_layers_list, target_layer, min_progress=0, max_progress=100,
        dissolve_output=None
):
    cfg.logger.log.debug('start')
    t_l = create_virtual_layer(input_layers_list)
    # open virtual layer
    input_source = ogr.Open(t_l)
    # get input layer definition
    input_layer = input_source.GetLayer()
    i_sr = input_layer.GetSpatialRef()
    input_layer_def = input_layer.GetLayerDefn()
    input_field_count = input_layer_def.GetFieldCount()
    i_d = ogr.GetDriverByName('GPKG')
    # create output
    output_source = i_d.CreateDataSource(target_layer)
    output_name = files_directories.file_name(target_layer)
    output_layer = output_source.CreateLayer(
        str(output_name), i_sr, ogr.wkbPolygon
    )
    # fields
    field_names = []
    for f in range(input_field_count):
        f_def = input_layer_def.GetFieldDefn(f)
        output_layer.CreateField(f_def)
        field_names.append(f_def.GetNameRef())
    if not dissolve_output:
        # add area field
        area_field_def = ogr.FieldDefn(cfg.area_field_name, ogr.OFTReal)
        area_field_def.SetWidth(30)
        area_field_def.SetPrecision(2)
        output_layer.CreateField(area_field_def)
    output_layer_def = output_layer.GetLayerDefn()
    # start copy
    layer_count = input_source.GetLayerCount()
    output_layer.StartTransaction()
    q = 0
    for i in input_source:
        q += 1
        cfg.progress.update(
            step=q, steps=layer_count, minimum=min_progress,
            maximum=max_progress, message='merging vectors',
            percentage=q / layer_count
        )
        i_name = i.GetName()
        i_layer = input_source.GetLayer(i_name)
        i_feature = i_layer.GetNextFeature()
        while i_feature:
            geometry = i_feature.GetGeometryRef()
            o_feature = ogr.Feature(output_layer_def)
            o_feature.SetGeometry(geometry)
            if not dissolve_output:
                # set area
                area = geometry.GetArea()
                o_feature.SetField(cfg.area_field_name, area)
            for c in range(input_field_count):
                try:
                    field_name = field_names[c]
                    field = i_feature.GetField(c)
                    o_feature.SetField(field_name, field)
                except Exception as err:
                    cfg.logger.log.error(err)
            output_layer.CreateFeature(o_feature)
            i_feature = i_layer.GetNextFeature()
    output_layer.CommitTransaction()
    cfg.logger.log.debug('target_layer: %s' % target_layer)
    return target_layer


# create virtual layer
def create_virtual_layer(input_layer_list, target_layer=None):
    source = '''
    <OGRVRTDataSource>
    '''
    for layer in input_layer_list:
        i = ogr.Open(layer)
        i_l = i.GetLayer()
        i_n = i_l.GetName()
        xml = '''
            <OGRVRTLayer name="%s">
                <SrcDataSource>%s</SrcDataSource>
                <SrcLayer>%s</SrcLayer>
            </OGRVRTLayer>
        '''
        source += xml % (i_n, layer, i_n)
    source += '''
    </OGRVRTDataSource>
    '''
    if target_layer is None:
        target_layer = cfg.temp.temporary_raster_path(extension=cfg.vrt_suffix)
    with open(target_layer, 'w') as file:
        file.write(source)
    return target_layer


# merge dissolve layer to new layer
def merge_dissolve_layer(
        input_layer, target_layer, column, y_list_coordinates, min_progress=0,
        max_progress=100
):
    cfg.logger.log.debug('start')
    # open virtual layer
    input_source = ogr.Open(input_layer)
    i_layer = input_source.GetLayer()
    i_layer_name = i_layer.GetName()
    i_layer_sr = i_layer.GetSpatialRef()
    i_layer_def = i_layer.GetLayerDefn()
    field_count = i_layer_def.GetFieldCount()
    i_d = ogr.GetDriverByName('GPKG')
    output_source = i_d.CreateDataSource(target_layer)
    output_name = files_directories.file_name(target_layer)
    o_layer = output_source.CreateLayer(
        str(output_name), i_layer_sr, ogr.wkbPolygon
    )
    # fields
    for f_c in range(field_count):
        field_def = i_layer_def.GetFieldDefn(f_c)
        o_layer.CreateField(field_def)
    # add area field
    area_field_def = ogr.FieldDefn(cfg.area_field_name, ogr.OFTReal)
    area_field_def.SetWidth(30)
    area_field_def.SetPrecision(2)
    o_layer.CreateField(area_field_def)
    o_layer_def = o_layer.GetLayerDefn()
    # get unique values
    sql = 'SELECT DISTINCT "%s" FROM "%s"' % (column, i_layer_name)
    unique_values = input_source.ExecuteSQL(sql, dialect='SQLITE')
    values = []
    id_list = []
    for i, f in enumerate(unique_values):
        values.append(f.GetField(0))
    # release sql results
    input_source.ReleaseResultSet(unique_values)
    sql_list = str(y_list_coordinates)[1:-1].replace("'", '')
    # for each value
    n = 0
    min_p = min_progress
    max_p = int((max_progress - min_progress) / (len(values) + 1))
    for v in values:
        n += 1
        cfg.progress.update(
            step=n, steps=len(values), minimum=min_p + max_p * (n - 1),
            maximum=min_p + max_p * n, message='dissolving polygons',
            percentage=n / len(values)
        )
        cfg.logger.log.debug('progress: %s' % str(n / len(values)))
        # get geometries to be dissolved
        sql = 'SELECT DISTINCT(ST_COLLECT(CastToMultiPolygon(geom))), ' \
              'GROUP_CONCAT(DISTINCT id) FROM (SELECT fid as ' \
              'id, geom FROM "%s" WHERE %s = %s) INNER JOIN (SELECT ' \
              'DISTINCT id FROM "rtree_%s_geom" WHERE miny IN (' \
              '%s) OR maxy IN (%s) ) USING (id)' % (
                  i_layer_name, column, str(v), i_layer_name, sql_list,
                  sql_list)
        unique_features = input_source.ExecuteSQL(sql, dialect='SQLITE')
        if unique_features is not None:
            uv_features = unique_features.GetNextFeature()
            field_value = uv_features.GetField(0)
            geometry_ref = uv_features.GetGeometryRef()
            if geometry_ref is not None:
                if not geometry_ref.IsValid():
                    # try to fix invalid geometry with buffer
                    geometry_ref = geometry_ref.Buffer(0.0)
                geometry_count = geometry_ref.GetGeometryCount()
                o_layer.StartTransaction()
                if geometry_count > 1:
                    for g in range(geometry_count):
                        g_geometry_ref = geometry_ref.GetGeometryRef(int(g))
                        try:
                            if g_geometry_ref is not None:
                                # try to fix invalid geometry with buffer
                                if not g_geometry_ref.IsValid():
                                    g_geometry_ref = g_geometry_ref.Buffer(0.0)
                                if g_geometry_ref.IsValid():
                                    o_feature = ogr.Feature(o_layer_def)
                                    o_feature.SetGeometry(g_geometry_ref)
                                    o_feature.SetField(column, v)
                                    # set area
                                    area = g_geometry_ref.GetArea()
                                    o_feature.SetField(
                                        cfg.area_field_name, area
                                    )
                                    o_layer.CreateFeature(o_feature)
                                else:
                                    # rollback
                                    o_layer.RollbackTransaction()
                                    o_layer.CommitTransaction()
                                    o_layer.StartTransaction()
                                    o_feature = ogr.Feature(o_layer_def)
                                    o_feature.SetGeometry(geometry_ref)
                                    o_feature.SetField(column, v)
                                    # set area
                                    area = geometry_ref.GetArea()
                                    o_feature.SetField(
                                        cfg.area_field_name, area
                                    )
                                    o_layer.CreateFeature(o_feature)
                                    break
                        except Exception as err:
                            str(err)
                            # rollback
                            o_layer.RollbackTransaction()
                            o_layer.CommitTransaction()
                            o_layer.StartTransaction()
                            o_feature = ogr.Feature(o_layer_def)
                            o_feature.SetGeometry(geometry_ref)
                            o_feature.SetField(column, v)
                            # set area
                            area = geometry_ref.GetArea()
                            o_feature.SetField(cfg.area_field_name, area)
                            o_layer.CreateFeature(o_feature)
                            break
                    cfg.logger.log.debug('added union geometries')
                # single geometry
                else:
                    o_feature = ogr.Feature(o_layer_def)
                    o_feature.SetGeometry(geometry_ref)
                    o_feature.SetField(column, v)
                    # set area
                    area = geometry_ref.GetArea()
                    o_feature.SetField(cfg.area_field_name, area)
                    o_layer.CreateFeature(o_feature)
                o_layer.CommitTransaction()
            if field_value is not None:
                id_list1 = field_value.split(',')
                id_list.extend(id_list1)
        # release sql results
        input_source.ReleaseResultSet(unique_features)
    cfg.progress.update(step=min_p + max_p * n, message='copying polygons')
    feature_count = i_layer.GetFeatureCount()
    # copy not dissolved features
    i_feature = i_layer.GetNextFeature()
    c = 0
    while i_feature:
        c += 1
        cfg.progress.update(
            step=c, steps=feature_count, minimum=min_p + max_p * n,
            maximum=max_progress, percentage=int(c / feature_count * 100)
        )
        i_fid = str(i_feature.GetFID())
        v_field = i_feature.GetField(column)
        if str(i_fid) not in id_list:
            i_geom = i_feature.GetGeometryRef()
            o_feature = ogr.Feature(o_layer_def)
            o_feature.SetGeometry(i_geom)
            o_feature.SetField(column, v_field)
            # set area
            area = i_geom.GetArea()
            o_feature.SetField(cfg.area_field_name, area)
            o_layer.CreateFeature(o_feature)
        i_feature = i_layer.GetNextFeature()
    cfg.logger.log.debug('end')
    return target_layer


# create a geometry vector for Spectral Signatures
def create_geometry_vector(
        output_path, crs_wkt, macroclass_field_name,
        class_field_name, vector_format='GPKG',
        ):
    # in case crs_wkt is already as crs format
    try:
        crs_wkt = str(crs_wkt.toWkt())
    except Exception as err:
        str(err)
    driver = ogr.GetDriverByName(vector_format)
    data_source = driver.CreateDataSource(output_path)
    spatial_ref = osr.SpatialReference()
    spatial_ref.ImportFromWkt(crs_wkt)
    name = files_directories.file_name(output_path, suffix=False)
    files_directories.create_parent_directory(output_path)
    layer = data_source.CreateLayer(name, spatial_ref, ogr.wkbMultiPolygon)
    # spectral signature id
    field_uid = ogr.FieldDefn(cfg.uid_field_name, ogr.OFTString)
    layer.CreateField(field_uid)
    macroclass_field = ogr.FieldDefn(macroclass_field_name, ogr.OFTInteger)
    layer.CreateField(macroclass_field)
    class_field = ogr.FieldDefn(class_field_name, ogr.OFTInteger)
    layer.CreateField(class_field)
    return output_path


# get polygon from vector and return memory layer
def get_polygon_from_vector(vector_path, attribute_filter=None):
    # open input vector
    vector = ogr.Open(vector_path)
    # get layer
    try:
        _v_layer = vector.GetLayer()
    except Exception as err:
        cfg.logger.log.error(err)
        return False
    # attribute filter
    _v_layer.SetAttributeFilter(attribute_filter)
    d = ogr.GetDriverByName('GPKG')
    temp = cfg.temp.temporary_file_path(name_suffix=cfg.gpkg_suffix)
    _d_s = d.CreateDataSource(temp)
    _d_s.CopyLayer(_v_layer, _v_layer.GetName(), ['OVERWRITE=YES'])
    _v_layer = None
    _d_s = None
    return temp



from unittest import TestCase

import remotior_sensus
from remotior_sensus.util import files_directories


class TestLog(TestCase):

    def test_create(self):
        rs = remotior_sensus.Session(
            n_processes=2, available_ram=1000, log_level=10
        )
        cfg = rs.configurations
        cfg.logger.log.debug('>>> test logger file path')
        self.assertTrue(files_directories.is_file(cfg.logger.file_path))

        # clear temporary directory
        rs.close()
